#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the sampling functions.
"""

import pytest
import itertools
import collections

from streprogen.sampling import sample_markov_ladder, sample_markov_loop


@pytest.mark.parametrize(
    "probabilities, structure",
    [
        ([1, 2, 3], [0.5, 0.5]),
        ([1, 2, 3], [0.7, 0.3]),
        ([1, 2, 3], [0.1, 0.9]),
        ([1, 2, 3, 7, 4], [0.2, 0.4, 0.4]),
        ([1, 2, 3, 7, 4], [0.5, 0.25, 0.25]),
        ([1, 2, 3, 7, 4], [1, 5]),
    ],
)
def test_markov_ladder(probabilities, structure):
    generator = sample_markov_ladder(probabilities, structure=structure)

    # Scale after calling sampling, to test that normalization is not needed
    probabilities = [p_i / sum(probabilities) for p_i in probabilities]

    # Draw samples and normalize output probabilities
    num_samples = 1_000_000
    samples = itertools.islice(generator, num_samples)
    counts = collections.Counter(samples)
    output_probabilities = [counts[k] / num_samples for k in sorted(counts.keys())]

    assert all((abs(p_i - y_i) / y_i) < 0.1 for p_i, y_i in zip(probabilities, output_probabilities))


@pytest.mark.parametrize(
    "probabilities, structure",
    [
        ([1, 2, 3], [0.5, 0.5]),
        ([1, 2, 3], [0.7, 0.3]),
        ([1, 2, 3], [0.1, 0.9]),
        ([1, 2, 3, 7, 4], [0.2, 0.4, 0.4]),
        ([1, 2, 3, 7, 4], [0.5, 0.25, 0.25]),
        ([1, 2, 3, 7, 4], [1, 5]),
    ],
)
def test_markov_loop(probabilities, structure):
    generator = sample_markov_loop(probabilities, structure=structure)

    # Scale after calling sampling, to test that normalization is not needed
    probabilities = [p_i / sum(probabilities) for p_i in probabilities]

    # Draw samples and normalize output probabilities
    num_samples = 1_000_000
    samples = itertools.islice(generator, num_samples)
    counts = collections.Counter(samples)
    output_probabilities = [counts[k] / num_samples for k in sorted(counts.keys())]

    assert all((abs(p_i - y_i) / y_i) < 0.1 for p_i, y_i in zip(probabilities, output_probabilities))


if __name__ == "__main__":
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[".", "--doctest-modules", "--capture=sys", "-v", "-k", "markov"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the modeling functions.
"""

import pytest

from streprogen import progression_sinh, progression_diffeq


@pytest.mark.parametrize("k, duration", [(0, 12), (0, 13), (3, 12), (3, 13)])
def test_progression_sinh(k, duration):
    # Test first and last points
    assert progression_sinh(1, 50, 100, 1, final_week=duration, k=k) == 50
    assert progression_sinh(duration, 50, 100, 1, final_week=duration, k=k) == 100

    # Test midpoint
    assert progression_sinh((duration + 1) / 2, 50, 100, 1, final_week=duration, k=k) == 75


@pytest.mark.parametrize("k, duration", [(0, 12), (0, 13), (3, 12), (3, 13)])
def test_progression_diffeq(k, duration):
    # Test first and last points
    assert progression_diffeq(1, 50, 100, 1, final_week=duration, k=k) == 50
    assert progression_diffeq(duration, 50, 100, 1, final_week=duration, k=k) == 100


if __name__ == "__main__":
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys", "-vv"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the API. The API should be considered fixed in future releases and should
be equal for every implementation. Therefore it's important to have unit tests
for the API.
"""

import pytest
import itertools
import statistics

from streprogen import Day, DynamicExercise, Program, StaticExercise
from streprogen import progression_sawtooth, progression_sinusoidal


@pytest.mark.parametrize("duration", list(range(2, 9)))
def test_verbose_api(duration):
    """The oldest, most verbose API."""

    def curl_func(week):
        return "{} x 10".format(week)

    # Create a 4-week program
    program = Program("My first program!", duration=duration)

    # Create some dynamic and static exercises
    bench = DynamicExercise("Bench press", 60, 60 * (1 + 0.01 * duration))
    squats = DynamicExercise("Squats", 80, 80 * (1 + 0.01 * duration))
    curls = StaticExercise("Curls", curl_func)

    day = Day()
    day.add_exercises(bench, squats, curls)

    assert len(day.exercises) == 3

    # Add day(s) to program and render it
    program.add_days(day)

    assert len(program.days) == 1

    assert not program._rendered
    program.render()
    assert program._rendered


def test_general_api():
    """The slightly more easy API."""

    def curl_func(week):
        return "{} x 10".format(week)

    # Create a 4-week program
    program = Program("My first program!", duration=8)

    # Create some dynamic and static exercises
    bench = DynamicExercise("Bench press", 60, 70)
    squats = DynamicExercise("Squats", 80, 90)
    curls = StaticExercise("Curls", curl_func)
    day = Day(exercises=[bench, squats, curls])

    # Variable reference to days are added
    assert bench.day is not None
    assert squats.day is not None

    assert len(day.exercises) == 3

    # Add day(s) to program and render it
    program.add_days(day)
    assert day.program is not None

    assert len(program.days) == 1

    assert not program._rendered
    program.render()
    assert program._rendered


def test_decorator_api():
    """The new API."""

    def curl_func(week):
        return "{} x 10".format(week)

    # Create a 4-week program
    program = Program("My first program!", duration=8)

    with program.Day():
        program.DynamicExercise("Bench press", 60, 65)
        program.DynamicExercise("Squats", 80, 85)
        program.StaticExercise("Curls", curl_func)

    assert len(program.days) == 1
    assert len(program.days[0].exercises) == 3

    assert program.days[0].program is not None
    assert program.days[0].exercises[0].day is not None

    assert not program._rendered
    program.render()
    assert program._rendered


@pytest.mark.parametrize("period", [0, 1])
def test_progression_periods(period):
    for week in range(1, 8 + 1):
        ans = progression_sawtooth(
            week,
            start_weight=1.0,
            final_weight=1.0,
            start_week=1,
            final_week=8,
            period=period,
            scale=1.0,  # Does not matter
            offset=0,  # Does not matter
            k=0,  # Does not matter
        )
        assert abs(ans - 1.0) < 1e-6

        ans = progression_sinusoidal(
            week,
            start_weight=1.0,
            final_weight=1.0,
            start_week=1,
            final_week=8,
            period=period,
            scale=1.0,  # Does not matter
            offset=0,  # Does not matter
            k=0,  # Does not matter
        )
        assert abs(ans - 1.0) < 1e-6


@pytest.mark.parametrize(
    "func, period",
    list(itertools.product([progression_sawtooth, progression_sinusoidal], [1, 2, 3, 4, 5])),
)
def test_progression_means(func, period):
    weeks = list(range(1, period + 1))
    target = 42

    values = [
        func(
            w,
            start_weight=target,
            final_weight=target,
            start_week=1,
            final_week=8,
            period=period,
            scale=1.0,  # Does not matter
            offset=1.0,  # Does not matter
            k=0,  # Does not matter
        )
        for w in weeks
    ]

    assert abs(statistics.mean(values) - target) <= 1e-6


def test_dynamic_exercises_are_not_mutated():
    """Creating a program should not mutate an exercise."""

    # Create an exercise
    exercise = DynamicExercise("Bench press", start_weight=100)
    exercise_dict = exercise.serialize()

    # Create a program, add the day and render it
    program = Program()
    day = Day("Monday")
    day.add_exercises(exercise)
    program.add_days(day)
    program.render()

    assert exercise.serialize() == exercise_dict


def test_program_not_mutated_after_rendering():
    """Creating and rendering a program should not mutate its dict
    representations."""

    # Set some non-typical parameters
    program = Program(
        name="MyProgram",
        duration=5,
        min_reps=1,
        reps_per_exercise=31,
        round_to=10,
        rep_scaler_func=[0.99, 0.97, 0.96, 0.95, 0.98],
        intensity_scaler_func=[0.99, 0.97, 0.96, 0.95, 0.98],
        units="asdf",
    )
    with program.Day("A"):
        program.DynamicExercise("Squats", start_weight=100, final_weight=113, max_reps=12)

    program_serialized = program.serialize()
    program.render()
    program_dict = program.to_dict()

    # Rendering the program should not change the serialization
    assert program.serialize() == program_serialized

    # Serializing and de-serializing should not change the program dict reprs
    program = Program.deserialize(program.serialize())
    program.render()
    assert program.serialize() == program_serialized
    assert program_dict == program.to_dict()


class TestProgressInformation:
    def test_progress_information_override_weights(self):
        program = Program(name="MyProgram", duration=10, percent_inc_per_week=10)
        with program.Day("A"):
            squat = program.DynamicExercise("Squats", start_weight=100, final_weight=150)

        (start_w, final_w, inc_week) = squat._progress_information()
        assert start_w == 100
        assert final_w == 150
        assert inc_week == 5  # Weight override program default

    def test_progress_information_override_perc_inc(self):
        program = Program(name="MyProgram", duration=10, percent_inc_per_week=1)
        with program.Day("A"):
            squat = program.DynamicExercise("Squats", start_weight=100, percent_inc_per_week=10)

        (start_w, final_w, inc_week) = squat._progress_information()
        assert start_w == 100
        assert final_w == 200
        assert inc_week == 10  # Weight override program default

    def test_progress_information_calcs(self):
        program = Program(name="MyProgram", duration=10, percent_inc_per_week=123)

        # Three ways of saying the same thing
        with program.Day():
            a = program.DynamicExercise("a", start_weight=100, percent_inc_per_week=10)
            b = program.DynamicExercise("b", start_weight=100, final_weight=200)
            c = program.DynamicExercise("c", final_weight=200, percent_inc_per_week=10)

        a_info = a._progress_information()
        b_info = b._progress_information()
        c_info = c._progress_information()

        assert a_info == b_info
        assert b_info == c_info

    def test_progress_information_calcs_from_program(self):
        program = Program(name="MyProgram", duration=10, percent_inc_per_week=10)

        # Three ways of saying the same thing
        with program.Day():
            a = program.DynamicExercise("a", start_weight=100)
            b = program.DynamicExercise("b", start_weight=100, final_weight=200)
            c = program.DynamicExercise("c", final_weight=200)

        a_info = a._progress_information()
        b_info = b._progress_information()
        c_info = c._progress_information()

        assert a_info == b_info
        assert b_info == c_info

    def test_progress_information_overspecified(self):
        # Set some non-typical parameters
        program = Program(name="MyProgram", duration=10)
        with pytest.raises(ValueError, match="At most 2 out of 3 variables may be set"):
            with program.Day("A"):
                program.DynamicExercise("Squats", start_weight=100, final_weight=150, percent_inc_per_week=10)


def test_error_on_non_unique_exercise_names():
    """Test that using the same exercise name raises an error."""

    program1 = Program(duration=8, round_to=1)
    with program1.Day():
        program1.DynamicExercise("Bench press", start_weight=100, min_reps=4, max_reps=7)
        program1.DynamicExercise("Bench press", start_weight=90, min_reps=1, max_reps=7)

    with pytest.raises(ValueError, match="Exercise name not unique: Bench press"):
        program1.render()


class TestSerialization:
    def test_DynamicExercise(self):
        """Serialize and deserialize should be equal."""

        dynamic_ex = DynamicExercise("Bench", start_weight=100)
        dyn_ex_dict = dynamic_ex.serialize()
        assert dynamic_ex == DynamicExercise.deserialize(dyn_ex_dict)

    def test_StaticExercise(self):
        """Serialize and deserialize should be equal."""

        static_ex = StaticExercise("Dips", "4 x 10")
        static_ex_dict = static_ex.serialize()
        assert static_ex == StaticExercise.deserialize(static_ex_dict)

    def test_Day(self):
        """Serialize and deserialize should be equal."""

        bench = DynamicExercise("Bench", start_weight=100)
        deadlift = DynamicExercise("Deadlift", start_weight=100, final_weight=140)
        static_ex = StaticExercise("Dips", "4 x 10")

        day = Day("Monday", [bench, deadlift, static_ex])
        day_dict = day.serialize()

        assert day_dict == Day.deserialize(day_dict).serialize()

    def test_Program(self):
        """Serialize and deserialize should be equal."""

        program = Program(
            name="Beginner 5x5", duration=4, intensity=85, units="kg", round_to=2.5, rep_scaler_func=[1, 1, 1, 1]
        )

        with program.Day("A"):
            program.DynamicExercise(name="Squat", start_weight=100)

        program.render()

        # Create a new program by serializing and deserializing
        new_program = Program.deserialize(program.serialize())
        new_program.render()

        assert str(program) == str(new_program)


class TestWaysOfGivingRepAndIntensity:
    @pytest.mark.parametrize("format", ["tex", "txt", "html", "dict"])
    def test_rep_scalers_as_function_vs_list(self, format):
        """Test that both functions and lists work the same."""

        def rep_scaler_func(week):
            return 1 - week / 100

        rep_scalers = [(1 - week / 100) for week in range(1, 9)]

        program1 = Program("My first program!", duration=8, round_to=1, rep_scaler_func=rep_scaler_func)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        program2 = Program("My first program!", duration=8, round_to=1, rep_scaler_func=rep_scalers)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)
        program2.render()

        # Use .txt format to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()

    @pytest.mark.parametrize("format", ["tex", "txt", "html", "dict"])
    def test_intensity_scalers_as_function_vs_list(self, format):
        """Test that both functions and lists work the same."""

        def intensity_scaler_func(week):
            return 1 - week / 100

        intensity_scalers = [(1 - week / 100) for week in range(1, 9)]

        program1 = Program("My first program!", duration=8, round_to=1, intensity_scaler_func=intensity_scaler_func)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        program2 = Program("My first program!", duration=8, round_to=1, intensity_scaler_func=intensity_scalers)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)
        program2.render()

        # Use .txt format to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()


class TestWaysOfGivingProgress:
    @pytest.mark.parametrize("format", ["tex", "txt", "html"])
    def test_rep_range_exercise_vs_program(self, format):
        """Test that giving rep range in program or exercise is the same."""

        program1 = Program("My first program!", duration=8, round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100, min_reps=4, max_reps=7)
        program1.render()

        program2 = Program("My first program!", duration=8, round_to=1, min_reps=4, max_reps=7)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)
        program2.render()

        # Use formats to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()

    @pytest.mark.parametrize("format", ["tex", "txt", "html"])
    def test_inc_week_program_vs_exercise(self, format):
        """Test that giving progress in program or exercise is the same."""

        program1 = Program(duration=8, round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100, percent_inc_per_week=2)
        program1.render()

        program2 = Program(duration=8, round_to=1, percent_inc_per_week=2)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)
        program2.render()

        # Use .txt format to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()

    @pytest.mark.parametrize("format", ["tex", "txt", "html"])
    def test_inc_per_week_vs_endpoints(self, format):
        """Test that giving progress in program or exercise is the same."""

        program1 = Program(duration=3, units="kg", round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100, percent_inc_per_week=2)
        program1.render()

        program2 = Program(duration=3, units="kg", round_to=1)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100, final_weight=106)
        program2.render()

        # Use .txt format to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()

    @pytest.mark.parametrize("format", ["tex", "txt", "html"])
    def test_start_weight_vs_final_weight(self, format):
        """Test that giving progress in program or exercise is the same."""

        program1 = Program(duration=3, units="kg", round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100, percent_inc_per_week=2)
        program1.render()

        program2 = Program(duration=3, units="kg", round_to=1)
        with program2.Day():
            program2.DynamicExercise("Bench press", percent_inc_per_week=2, final_weight=106)
        program2.render()

        # Use .txt format to compare programs
        assert getattr(program1, f"to_{format}")() == getattr(program2, f"to_{format}")()

    def test_setting_optimization_params(self):
        """Test that setting via method works."""

        program1 = Program(duration=12, units="kg", round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        program2 = Program(duration=12, units="kg", round_to=1)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)

        program2.set_optimization_params(reps_slack=3, max_diff=1, max_unique=2)
        program2.render()

        # They should be different
        assert str(program1) != str(program2)


class TestShiftingDynExercises:
    def test_shifting_zero(self):
        program1 = Program("My first program!", duration=8, round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        program2 = Program("My first program!", duration=8, round_to=1)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100, shift=0)
        program2.render()

        assert str(program1) == str(program2)

    @pytest.mark.parametrize("shift", [-3, -2, -1, 0, 1, 2, 3])
    def test_shifting_k(self, shift):
        program1 = Program("My first program!", duration=8, round_to=1)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        program2 = Program("My first program!", duration=8, round_to=1)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100, shift=shift)
        program2.render()

        assert program1.to_dict()["rendered"][4] == program2.to_dict()["rendered"][4 - shift]

    def test_list_vs_function_arg(self):
        """Test that index 0 in the list corresponds to week=1."""

        def rep_scaler_func(week):
            return 1 + week / 10

        program1 = Program("My first program!", duration=8, rep_scaler_func=rep_scaler_func)
        with program1.Day():
            program1.DynamicExercise("Bench press", start_weight=100)
        program1.render()

        rep_scaler_list = [1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8]

        program2 = Program("My first program!", duration=8, rep_scaler_func=rep_scaler_list)
        with program2.Day():
            program2.DynamicExercise("Bench press", start_weight=100)
        program2.render()

        assert str(program1) == str(program2)


class TestRounding:
    def test_rounding_func_exercise(self):
        """Test that both functions and numbers work for rounding."""

        def round_function(x):
            return round(x / 4) * 4

        # Create a 4-week program
        program = Program("My first program!", duration=8)

        with program.Day():
            program.DynamicExercise("Bench press", 60, 65, round_to=2.5)
            program.DynamicExercise("Squats", 80, 85, round_to=round_function)

        program.render()

        assert True

    def test_rounding_func_program(self):
        """Test that both functions and numbers work for rounding."""

        def round_function(x):
            return round(x / 4) * 4

        # Create a 4-week program
        program = Program("My first program!", duration=8, round_to=round_function)

        with program.Day():
            program.DynamicExercise("Bench press", 60, 65)
            program.DynamicExercise("Squats", 80, 85)

        program.render()

        for week in program.to_dict()["rendered"]:
            for day in week:
                for dyn_ex in day["exercises"]:
                    assert all(w % 4 == 0 for w in dyn_ex["weights"])

        assert True


if __name__ == "__main__":
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[".", "--doctest-modules", "--capture=sys", "-v", "-k", "round"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numbers
import itertools
import random
import bisect


def roll(container, k):
    """Roll elements in a container, wrapping around the boundary.

    Examples
    --------
    >>> x = list(range(10))
    >>> roll(x, 2)
    [8, 9, 0, 1, 2, 3, 4, 5, 6, 7]

    """
    size = len(container)
    return [container[(i - k) % size] for i in range(size)]


def sample(weights):
    """Yield integers corresponding to weighted samples taken with replacement."""
    weights = list(weights)
    assert all(isinstance(w_i, numbers.Number) for w_i in weights)
    assert all(w_i >= 0 for w_i in weights)

    cumsum = list(itertools.accumulate(weights))
    while True:
        pick = random.random() * cumsum[-1]
        yield bisect.bisect_left(cumsum, x=pick)


def sample_markov_loop(probabilities, structure=None):
    """Sample from a probability distribution over 0, 1, ..., n-1 given by
    a sequence of `probabilities`. The samples are generated using a Markov
    chain with structure that allows jumping forward.

    The first sample will always be 0. To start at a random state, run the
    chain for a while and discard the beginning of it.

    The `structure` argument determines both how many steps ahead it is
    possible to sample, and what relative preference is given to the steps.

    Examples
    --------
    >>> probabilities = [1, 2, 3, 4]
    >>> structure = [0.5, 0.5] # Either stay in state i, or jump to i+1
    >>> structure = [0.1, 0.9] # Same, but the chain mixes faster
    >>> structure = [1, 0] # Same, but edge case - chain never mixes
    >>>
    >>> structure = [0, 0.8, 0.2] # Allows jumping 1 or 2 states ahead
    >>> structure = [0, 0.1, 0.9] # Same, but jumps 2 states more often

    """
    # If no structure is given, then strongly prefer jumping one step ahead
    if structure is None:
        structure = [0.01, 0.99]

    # The structure can at most have the same length as the probability vector
    assert len(structure) <= len(probabilities)
    probabilities, structure = list(probabilities), list(structure)

    # Convert to lists just to be safe, then pad the structure with zeros
    structure = structure + [0] * (len(probabilities) - len(structure))

    # Normalize structure, then interpret it relative to min(probabilities)
    # We must have 0 <= s_1 + s_2 + ... <= min(probability)
    min_probability = min(probabilities)
    sum_structure = sum(structure)
    structure = [min_probability * s_i / sum_structure for s_i in structure]

    # On the diagonals we find S = s_1 + s_2 + s_3 + ..., and we store this
    # value of S on the 0'th index for convenience
    structure[0] = sum(structure[1:])

    # Start at state 0. To start at a random state, throw away initial samples
    state = 0
    yield state

    while True:
        # Compute the row in the transition matrix corresponding to the state
        to_draw_probs = roll(structure, state)
        # In the diagonal entry, we must subtract. P_ii = pi_i - S
        to_draw_probs[state] = probabilities[state] - to_draw_probs[state]
        # Divide every element by pi_i
        to_draw_probs = [p_i / probabilities[state] for p_i in to_draw_probs]

        state = next(sample(to_draw_probs))
        yield state


def sample_markov_ladder(probabilities, structure=None):
    """Sample from a probability distribution over 0, 1, ..., n-1 given by
    a sequence of `probabilities`. The samples are generated using a Markov
    chain with structure that allows jumping to neighboring states (but does
    NOT loop around).

    The first sample will always be 0. To start at a random state, run the
    chain for a while and discard the beginning of it.

    The `structure` argument determines both how many neighboring states it is
    possible to jump to, and what relative preference is given to the jumps.

    Examples
    --------
    >>> probabilities = [1, 2, 3, 4]
    >>> structure = [0.5, 0.5] # Either stay in state i, or jump to i-1 or i+1
    >>> structure = [0.1, 0.9] # Same, but the chain mixes faster
    >>> structure = [1, 0] # Same, but edge case - chain never mixes
    >>>
    >>> structure = [0, 0.8, 0.2] # Allows jumping 1 or 2 states back or forth
    >>> structure = [0, 0.1, 0.9] # Same, but jumps further away more often

    """
    if structure is None:
        structure = [0.01, 0.99]

    assert len(structure) <= len(probabilities)
    probabilities, structure = list(probabilities), list(structure)

    # Normalize structure
    sum_structure = sum(structure)
    min_probability = min(probabilities)
    structure = [min_probability * (s_i / sum_structure) / 2 for s_i in structure]

    # Future states
    future_states = list(range(len(probabilities)))

    structure[0] = sum(structure[1:])

    state = 0
    yield state

    k, n = len(structure), len(probabilities)
    while True:
        # Assemble probabilities
        left_probs = [structure[i] / probabilities[state] for i in reversed(range(1, min(k, state + 1)))]
        right_probs = [structure[i] / probabilities[state] for i in range(1, min(k, n - state))]
        center_prob = [1 - sum(left_probs) - sum(right_probs)]
        to_draw_probs = left_probs + center_prob + right_probs

        future_states_filtered = [s for s in future_states if abs(state - s) < k]

        # Draw the index of the weight, then map it to the state integer
        state_index = next(sample(to_draw_probs))
        state = future_states_filtered[state_index]

        yield state


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys", "-k", "markov"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Data classes for optimizing meals.
"""
import warnings

# =============================================================================
# CLASSES - Used to store data and ease computations in hierarchical data
# =============================================================================


class Food:
    """
    A food consists of a name and nutritional data given in units of 100 grams.

    Examples
    --------
    >>> eggs = Food(name='eggs', protein=13.0, fat=10.6, carbs=0.3, kcal=149,
    ...             price_per_product=32.9, grams_per_product=690)
    >>> eggs.price
    4.768115942028985
    >>> eggs.name
    'eggs'
    >>> eggs.kcal
    149
    """

    def __init__(self, name, protein, fat, carbs, kcal, price_per_product, grams_per_product):
        """TODO"""
        self.name = name
        self.protein = protein
        self.fat = fat
        self.carbs = carbs
        self.kcal = kcal
        self.price_per_product = price_per_product
        self.grams_per_product = grams_per_product

        self._verify()

    @property
    def price(self):
        return self.price_per_product / self.grams_per_product * 100

    def _verify(self):
        """Verify the relationship between macros and kcal."""
        computed_kcal = 4 * self.protein + 4 * self.carbs + 9 * self.fat
        relative_error = abs((self.kcal - computed_kcal) / computed_kcal)
        if relative_error > 0.1:
            msg = "Got a {} % error on kcal: '{}'."
            warnings.warn(msg.format(round(relative_error * 100, 1), self.name))

    def __repr__(self):
        name = type(self).__name__
        args = ("{}={}".format(arg, value) for (arg, value) in self.__dict__.items())
        args = ", ".join(args)

        return name + "({})".format(args)

    def __hash__(self):
        return hash((self.name, self.protein, self.fat, self.carbs, self.kcal))


class Meal:
    """
    A meal consists of several foods given at some "base" unit of grams. In the example
    below 'eggs' is the general food, an a meal consists of discrete units of eggs,
    weighing approximately 65 grams each.

    Examples
    --------
    >>> eggs = Food(name='eggs', protein=13.0, fat=10.6, carbs=0.3, kcal=149,
    ...             price_per_product=32.9, grams_per_product=690)
    >>> egg = Meal(name='egg', foods={eggs:65}, discrete=True)
    >>> egg.price # price of the meal
    3.0992753623188407
    """

    # Foods are added as: foods={all_foods["lettmelk"]:100, all_foods["musli"]:100}
    # This means that a baseline

    def __init__(self, name, foods, discrete=True):
        self.name = name
        self.foods = foods
        self.discrete = discrete
        if self.kcal < 10:
            warnings.warn("Food only has {} calories.".format(self.kcal))

    def __getattr__(self, key):
        """Allow accessing attributes of foods, summing over them."""
        return sum(getattr(food, key) * quantity / 100 for (food, quantity) in self.foods.items())

    @property
    def grams(self):
        return int(sum(self.foods.values()))

    def __hash__(self):
        return hash(self.name) + hash(frozenset(self.foods.keys()))

    def __iter__(self):
        return iter(self.foods.keys())

    def __copy__(self):
        return type(self)(name=self.name, foods=self.foods.copy(), discrete=self.discrete)

    def __str__(self):
        name = type(self).__name__
        foods_names = (
            "{" + ", ".join("{}g {}".format(quantity, food.name) for food, quantity in self.foods.items()) + "}"
        )
        return name + "(name='{}', grams={}, foods={})".format(self.name, self.grams, foods_names)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from streprogen.day import Day
from streprogen.diet import Food, Meal
from streprogen.exercises import DynamicExercise, StaticExercise
from streprogen.mealplan import Mealplan
from streprogen.modeling import (
    progression_diffeq,
    progression_sinh,
    progression_sinusoidal,
    progression_sawtooth,
    reps_to_intensity,
    reps_to_intensity_relaxed,
    reps_to_intensity_tight,
)
from streprogen.optimization import RepSchemeGenerator, RepSchemeOptimizer
from streprogen.program import Program
from streprogen.sampling import sample_markov_ladder, sample_markov_loop

__all__ = [
    "StaticExercise",
    "DynamicExercise",
    "Day",
    "Program",
    "reps_to_intensity",
    "reps_to_intensity_tight",
    "reps_to_intensity_relaxed",
    "progression_diffeq",
    "progression_sinh",
    "progression_sinusoidal",
    "progression_sawtooth",
    "Food",
    "Meal",
    "Mealplan",
    "RepSchemeGenerator",
    "RepSchemeOptimizer",
    "sample_markov_ladder",
    "sample_markov_loop",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import statistics
import numbers

from ortools.linear_solver import pywraplp


class RepSchemeGenerator:
    def __init__(self, reps_slack: int = 3, max_diff: int = 1, max_unique: int = 3):
        """Initialize the generator by supplying hard constraints.

        Parameters
        ----------
        reps_slack : int, optional
            Maximum deviation from the repetition goal.
        max_diff : int, optional
            Maximum difference between two consecutive sets.
        max_unique : int, optional
            Maximum unique sets in the solution.

        Examples
        --------
        >>> generator = RepSchemeGenerator(reps_slack=0, max_diff=2)
        >>> for result in generator.generate(sets=[2, 3, 4], reps_goal=6):
        ...     print(result)
        (2, 2, 2)
        (2, 4)
        (3, 3)
        >>> generator = RepSchemeGenerator(reps_slack=1, max_diff=2)
        >>> for result in generator.generate(sets=[2, 3, 4], reps_goal=6):
        ...     print(result)
        (2, 2, 2)
        (2, 2, 3)
        (2, 3)
        (2, 4)
        (3, 3)
        (3, 4)
        >>> generator = RepSchemeGenerator(reps_slack=0, max_diff=1, max_unique=1)
        >>> for result in generator.generate(sets=[2, 3, 4], reps_goal=8):
        ...     print(result)
        (2, 2, 2, 2)
        (4, 4)

        """
        assert isinstance(reps_slack, numbers.Integral)
        assert reps_slack >= 0
        assert isinstance(max_diff, numbers.Integral)
        assert max_diff >= 0
        assert isinstance(max_unique, numbers.Integral)
        assert max_unique >= 1

        self.reps_slack = reps_slack
        self.max_diff = max_diff
        self.max_unique = max_unique

    def generate(self, sets: list, reps_goal: int):
        """
        Generate repetition schemes.

        Parameters
        ----------
        sets : list
            A list of allowed repetitions. Unique and sorted.
        reps_goal : int
            The repetition goal.

        Yields
        ------
        Tuples with repetition schemes.

        """
        assert isinstance(reps_goal, numbers.Integral)
        assert reps_goal >= 0
        assert all(s > 0 for s in sets)

        # Store parameters in instance to avoid passing to _generate_sets()
        self.sets = list(sorted(set(sets)))
        self.reps_goal = reps_goal

        # Generate sets and yield them out
        for scheme in self._generate_sets(stack=[]):
            yield scheme

    def _generate_sets(self, i: int = 0, stack=None):
        """Only to be called internally."""
        assert stack is not None, "'stack' should be set to [] by caller."
        assert i >= 0
        assert isinstance(i, numbers.Integral)

        # Prune solutions with too many unique repetitions
        if len(set(stack)) > self.max_unique:
            return

        # Yield the result if it's within the allowed range
        if stack and (abs(sum(stack) - self.reps_goal) <= self.reps_slack):
            yield tuple(stack)

        # Stop the recursion if the sum is too high. This prunes the search.
        if sum(stack) > self.reps_goal + self.reps_slack:
            return

        for j in range(i, len(self.sets)):
            if not stack:
                yield from self._generate_sets(i=j, stack=stack + [self.sets[j]])
            else:
                # Prune solutions with too large differences
                # This avoids jumps like e.g. [8, 8, 3, 3]
                if self.sets[j] - stack[-1] <= self.max_diff:
                    yield from self._generate_sets(i=j, stack=stack + [self.sets[j]])


class RepSchemeOptimizer:
    def __init__(self, generator=None):
        """Initialize the optimizer.

        Parameters
        ----------
        generator : callable, optional
            A rep scheme generator with a 'generate' method with signature
            generate(self, sets: list, reps_goal: int).
            The default when `None` is the RepSchemeGenerator.

        """
        if generator is None:
            self.generator = RepSchemeGenerator()
        else:
            self.generator = generator

        self._cache = dict()

    def _optimize(self, sets: tuple, intensities: tuple, reps_goal: int, intensity_goal: float):
        """Core optimization. Moved to its own method for caching."""

        # Convert data to lists (tuples are used for caching)
        sets = list(sets)
        intensities = list(intensities)

        # Create a mapping
        reps_to_intensity_dict = {r_j: i_j for (r_j, i_j) in zip(sets, intensities)}

        def reps_to_intensity(reps):
            """Map repetitions to intensity, e.g. 8 -> 70."""
            return reps_to_intensity_dict[reps]

        schemes = self.generator.generate(sets=sets, reps_goal=reps_goal)

        def loss(scheme: list):
            """Loss function - smaller is better. Uses 2-norm."""
            reps = sum(scheme)
            intensities = map(reps_to_intensity, scheme)
            intensity = sum(r * i for r, i in zip(scheme, intensities)) / reps
            # No weighting is used here since the units are about the same
            return (reps - reps_goal) ** 2 + (intensity - intensity_goal) ** 2

        # Optimization here is simply going through every feasible solution
        # and choosing the best one. The feasible solution search space is so
        # small that this is fast and efficient, provided that the generator
        # is efficient.
        return list(reversed(min(schemes, key=loss)))

    def __call__(self, sets: tuple, intensities: tuple, reps_goal: int, intensity_goal: float):
        """Use the generator to generate feasible solutions, then optimize."""
        assert isinstance(sets, tuple)
        assert isinstance(intensities, tuple)
        assert reps_goal > 0
        assert intensity_goal > 1
        assert all(i_j > 1 for i_j in intensities)
        assert list(sets) == sorted(sets)

        # Pack the arguments into a tuple for caching
        args = (sets, intensities, reps_goal, intensity_goal)

        # Try to hit the cache. If it fails: compute, store and return.
        try:
            return self._cache[args]
        except KeyError:
            self._cache[args] = self._optimize(*args)
            return self._cache[args]


@functools.lru_cache(maxsize=1024, typed=False)
def optimize_sets(reps, intensities, reps_goal, intensities_goal):
    r"""Get the optimal number of sets.

    Introduction
    ------------

    This function uses a MIP model to optimize the number of sets in an exercise in a
    strength training program. Novices usually train with SETS x REPS, for instance
    3 x 12 - three sets of twelwe repetitions for some exercise. This is boring and does
    not allow for long-term planning.

    Consider a set of avaiable repetitions, {r_j}, and a set of corresponding
    intensities {i_j}. A realistic table might look like:

        reps          8     7      6      5      4      3
        intensities   0.7   0.74   0.78   0.82   0.86   0.90

    This means that this athlete can perform 6 repetitions at 78% of his/her maximum.
    From these sets we would like to draw reps and intensities to create programs, e.g.,

        8 x 0.7   -   6 x 0.78   -   6 x 0.78   -   5 x 0.82

    This corresponds to 25 repetitions and a (weighted) average intensity of
    (8 * 0.7 + 6 * 0.78 + 6 * 0.78 + 5 * 0.82) / 25 = 0.7624

    Mathematical model
    ------------------
    Let x_j be the number of sets performed with r_j repetitions, and a assume we have
    intensities i_j computed. We use goal programming to balance three constrains:

        (1) Number of reps.
            We want \sum_j x_j * r_j = reps_goal. In other words, the total number of
            reps chosen should be close to the goal. We will replace equality with slack
            variables using the goal progamming techniquer.

        (2) Average intensity.
            We want (\sum_j x_j * r_j * i_j) / (\sum_j x_j * r_j) = intensities_goal.
            We rewrite to: (\sum_j x_j * r_j * i_j) = intensities_goal * (\sum_j x_j * r_j)
            Then we use slack variables and goal programming to get close.

        (3) Dense solutions.
            We do not want solutions such as x_8 = 2 and x_3 = 4 and every other
            component of x to be zero. The jump from 8 reps to 3 reps is too high for
            the athlete. The first measure taken to ensure this is to maximize the
            density of the solution. We define z_j as

                z_j = 1 if x_j >= 1, else z_j = 0

            Which is achieved by adding the following constraints for every j:

                z_j <= x_j <= M z_j
                M = upper bound on x_j
                z_j \in {0, 1} (binary integer variable)

            And then we add -\sum_j z_j to the minimization problem.

        (4) Solutions with a small range (i.e. maximum - minimum).
            With the three constrains above, solutions tend to almost always use the
            entire range, going from .e.g 8 reps to 1 reps very often. It is not
            desireable to use the entire range of possible repetitions in every workout.

            We minimize the range by defining variables `minimum` and `maximum` and
            adding the constrains

                maximum >= z_j * r_j
                minimum <= z_j * r_j + (1 - z_j) * M
                M = upper bound on z_j * r_j

            The second constraint is interpreted as "`minimum` is smaller than z_j * r_j
            when z_j = 1 (when x_j >= 1), when z_j = 0 `minimum` is smaller than M".
            Then we add `maximum` - `minimum` to the objective, which we minimize.

    We use goal programming on all of these objectives, and we normalize them in a
    sensible way so a user can weight them as he pleases.

    Example
    -------
    >>> reps = tuple([8, 7, 6, 5, 4, 3, 2, 1])
    >>> intensities = tuple([0.7, 0.74, 0.78, 0.82, 0.86, 0.9, 0.94, 0.98])
    >>> reps_goal = 25
    >>> intensities_goal = 0.8
    >>> x, data = optimize_sets(reps, intensities, reps_goal, intensities_goal)
    """
    assert isinstance(reps, tuple)
    assert isinstance(intensities, tuple)
    assert len(reps) == len(intensities)
    assert all(r_j >= 1 for r_j in reps)
    assert all(0 <= i_j <= 1 for i_j in intensities)

    if max(intensities) < intensities_goal:
        # =============================================================================
        #         msg = (
        #             "Cannot reach goal intensity {} with intensities {}. Recasting goal value."
        #         )
        #         warnings.warn(msg.format(intensities_goal, intensities))
        # =============================================================================
        return optimize_sets(reps, intensities, reps_goal, intensities_goal=max(intensities))

    if min(intensities) > intensities_goal:
        # =============================================================================
        #         msg = (
        #             "Cannot reach goal intensity {} with intensities {}. Recasting goal value."
        #         )
        #         warnings.warn(msg.format(intensities_goal, intensities))
        # =============================================================================
        return optimize_sets(reps, intensities, reps_goal, intensities_goal=min(intensities))

    # The loss measure are normalized in the code, so the ratio of these values
    # will prioritize the goals relatively to each other.
    # If a value is set to 0, the solver will not care about it
    penalty_reps = 2.0
    penalty_intensity = 1.0
    reward_density = 1.0
    penality_spread = 0.9

    # =============================================================================
    #     SET UP PROBLEM, SOLVER AND DECISION VARIABLES
    # =============================================================================

    # Create a solver and an objective function
    solver = pywraplp.Solver("solver", pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)
    objective_function = 0
    INF = solver.infinity()

    # Create decision variables. x_j = number of sets with j reptitions
    x = [solver.IntVar(0, INF, "x_{}".format(r_j)) for r_j in reps]
    # Must choose at least one set
    solver.Add(sum(x) >= 1)

    # =============================================================================
    #     GOAL 1 : REPS AS CLOSE AS POSSIBLE TO DESIRED
    # =============================================================================

    # Set repetition goals
    total_reps = sum(x_j * r_j for (x_j, r_j) in zip(x, reps))

    # Add many linear penalities with increasing 'diff' to create a convex penalty
    for diff in range(10):
        delta = solver.NumVar(0, INF, "slack_reps_{}".format(diff))
        solver.Add(total_reps - reps_goal <= delta + diff)
        solver.Add(reps_goal - total_reps <= delta + diff)
        objective_function += penalty_reps * (delta) / reps_goal

    # =============================================================================
    #     GOAL 2 : INTENSITY AS CLOSE AS POSSIBLE TO DESIRED
    # =============================================================================

    numerator = sum(r_j * x_j * i_j for (r_j, x_j, i_j) in zip(reps, x, intensities))
    denominator = sum(r_j * x_j for (r_j, x_j) in zip(reps, x))

    # Add many linear penalities with increasing 'diff' to create a convex penalty
    for diff in range(10):
        delta = solver.NumVar(0, INF, "slack_intensity_{}".format(diff))
        solver.Add(numerator - intensities_goal * denominator <= delta + diff)
        solver.Add(intensities_goal * denominator - numerator <= delta + diff)
        objective_function += penalty_intensity * (delta) / intensities_goal

    # =============================================================================
    #     GOAL 3: AS DENSE SOLUTION AS POSSIBLE
    # =============================================================================

    # Create binary indicator variables z_j
    # z_j = 1 if and only if x_j >= 1, if not then z_j = 0
    M = 2 * reps_goal
    z = [solver.IntVar(0, 1, "z_{}".format(r_j)) for r_j in reps]
    for x_j, z_j in zip(x, z):
        solver.Add(z_j <= x_j)
        solver.Add(x_j <= M * z_j)

    # Maximize the density of the solution vector
    objective_function -= reward_density * sum(z) / len(intensities)

    # =============================================================================
    #     GOAL 4: AS LITTLE SPREAD (MAX - MIN) AS POSSIBLE
    # =============================================================================

    minimum = solver.NumVar(0, INF, "minimum_reps")
    maximum = solver.NumVar(0, INF, "maximum_reps")
    # This is to prevent numerical issues
    solver.Add(minimum <= maximum)

    for r_j, z_j in zip(reps, z):
        solver.Add(z_j * r_j <= maximum)
        solver.Add(z_j * r_j + (1 - z_j) * M >= minimum)

    denom = max(max(reps) - min(reps), 0.0001)
    objective_function += penality_spread * (maximum - minimum) / denom

    # =============================================================================
    #     SOLVE THE PROBLEM, PARSE THE SOLUTION AND RETURN
    # =============================================================================

    # Minimize the deviation from the goal
    solver.Minimize(objective_function)
    result_status = solver.Solve()

    assert result_status == pywraplp.Solver.OPTIMAL
    assert solver.VerifySolution(1e-7, True)

    x = [x_j.solution_value() for x_j in x]
    z = [z_j.solution_value() for z_j in z]

    # Assert that the property of the z's hold
    for x_j, z_j in zip(x, z):
        if z_j == 1:
            assert x_j >= 1
        else:
            assert x_j == 0

    numerator = sum(r_j * x_j * i_j for (r_j, x_j, i_j) in zip(reps, x, intensities))
    denominator = sum(r_j * x_j for (r_j, x_j) in zip(reps, x))

    return (
        x,
        {
            "intensity": round(numerator / denominator, 4),
            "reps": denominator,
            "obj_func_value": round(solver.Objective().Value(), 4),
            "minimum": minimum.solution_value(),
            "maximum": maximum.solution_value(),
        },
    )


def optimize_mealplan(
    meals,
    dietary_constraints,
    meals_limits=None,
    num_days=1,
    num_meals=4,
    time_limit_secs=10,
    epsilon=1e-3,
    weight_price=0.1,
    weight_nutrients=2.0,
    weight_range=0.75,
    params=None,
):
    """Optimize the quantitiy of each meal in a day, given constraints."""

    # =============================================================================
    #     PARSE INPUT ARGUMENTS
    # =============================================================================

    assert isinstance(meals, (list, tuple))
    assert isinstance(dietary_constraints, (dict,))
    assert isinstance(params, (dict,)) or params is None
    assert (meals_limits is None) or isinstance(meals_limits, (list, tuple))

    meals = meals.copy()
    dietary_constraints = dietary_constraints.copy()

    if meals_limits is None:
        meals_limits = [(None, None) for meal in meals]

    if params is None:
        params = dict()

    # A small number such as 0.001. x_ij >= EPSILON <=> z_ij = 1
    EPSILON = epsilon

    expected_daily_price = params.get("expected_daily_price", 75)
    M1 = params.get("M1", 50)  # Upper bound on x_ij
    M2 = params.get("M2", 50)  # Upper bound on x[i][j] * meal.kcal, i.e. calories in a meal

    # A strange bug is that sometime the optimizer will return INFEASIBLE on attempt #1,
    # but calling this function again with the same inputs works. So we allow calling it
    # two times. This is very hacky and not suitable for production, but it works for
    # 'hobby usage'.
    first_call = params.get("first_call", True)

    # Create a solver and an objective function
    solver_name = str(hash(tuple(meals)) + hash((num_days, num_meals)))
    solver = pywraplp.Solver(solver_name, pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)
    solver.set_time_limit(time_limit_secs * 1000)
    objective_function = 0
    INF = solver.infinity()

    # =============================================================================
    #     ERROR CHECKING AND USER INPUT SANITATION
    # =============================================================================

    # Error checking
    meals_total = num_days * num_meals
    maximum_limit = sum(9999 if high is None else high for (low, high) in meals_limits)
    if maximum_limit < meals_total:
        msg = "Cannot achieve {} totals meals with a total of {} meals."
        raise RuntimeError(msg.format(meals_total, maximum_limit))

    allowed_macros = ("kcal", "protein", "fat", "carbs")
    assert all(key in allowed_macros for key in dietary_constraints.keys())

    assert len(meals_limits) == len(meals)

    # =============================================================================
    #     CREATE VARIABLES
    # =============================================================================

    x = [[None for j in range(num_days)] for i in range(len(meals))]
    z = [[None for j in range(num_days)] for i in range(len(meals))]

    # Loop over every combination of meal and days, create variables
    for i, meal in enumerate(meals):
        for j in range(num_days):
            z[i][j] = solver.IntVar(0, 1, "z_{}{}".format(i, j))

            if meal.discrete:
                x[i][j] = solver.IntVar(0, INF, "x_{}{}".format(i, j))
            else:
                x[i][j] = solver.NumVar(0, INF, "x_{}{}".format(i, j))

            # These constraints ensure that z_ij = 1 iff x_ij >= EPSILON
            solver.Add(EPSILON * z[i][j] <= x[i][j])
            eps = EPSILON / 10
            solver.Add(x[i][j] <= (M1 + eps) * z[i][j] + EPSILON - eps)

    # =============================================================================
    #     CREATE CONSTRAINTS / OBJECTIVE FUNCTION TERM
    # =============================================================================

    # OBJECTIVE FUNCTION TERM 1: Total price of the meals in the program
    denom = expected_daily_price * num_days
    for j in range(num_days):
        daily_price = sum(x[i][j] * meal.price for i, meal in enumerate(meals))
        objective_function += (weight_price / denom) * daily_price

    # OBJECTIVE FUNCTION TERM 2: Deviation from nutrients (on a daily basis)
    for j in range(num_days):
        for macro, (low, high) in dietary_constraints.items():
            # No point in adding any constraints if it's None
            if low is None and high is None:
                continue

            food_macros = [getattr(meal, macro) for meal in meals]

            # Create the sum: sum_i food_i * macro_i
            x_meals = [x[i][j] for i in range(len(meals))]
            total_macro = sum(c * x for x, c in zip(x_meals, food_macros))

            # The maximal deviation in a day is approx mean([low, high]) * nutrients
            # The maximal deviation is the above times the number of days
            denom = statistics.mean([value for value in [low, high] if value is not None])
            denom = denom * num_days  # * len(dietary_constraints)

            # Slack variables related to the lower limit. Only "undershooting" is penalized.
            if low is not None:
                low_positive = solver.NumVar(0, INF, "over_low_lim_" + macro + str(j))
                low_negative = solver.NumVar(0, INF, "under_low_lim_" + macro + str(j))
                solver.Add(total_macro + low_positive - low_negative == low)
                objective_function += (weight_nutrients / denom) * low_positive

            # Slack variables related to the upper limit. Only "overshooting" is penalized.
            if high is not None:
                high_positive = solver.NumVar(0, INF, "over_up_lim_" + macro + str(j))
                high_negative = solver.NumVar(0, INF, "under_upp_lim_" + macro + str(j))
                solver.Add(total_macro + high_positive - high_negative == high)
                objective_function += (weight_nutrients / denom) * high_negative

    # OBJECTIVE FUNCTION TERM 3: Minimal range on calories (on a daily basis)
    for j in range(num_days):
        lower = solver.NumVar(0, INF, "lower_kcal_{}".format(j))
        upper = solver.NumVar(0, INF, "upper_kcal_{}".format(j))

        for i, meal in enumerate(meals):
            solver.Add(lower <= x[i][j] * meal.kcal + (1 - z[i][j]) * M2)
            solver.Add(upper >= x[i][j] * meal.kcal)

        # The maximal spread per day is approximately mean([kcal_low, kcal_high]) / meals
        # The maximal spread is the above times the number of days. Normalize w.r.t this
        denom = statistics.mean([value for value in dietary_constraints["kcal"] if value is not None])
        denom = denom * num_days / num_meals
        objective_function += (weight_range / denom) * (upper - lower)

    # HARD CONSTRAINT 1 : Number of meals per day
    for j in range(num_days):
        solver.Add(sum(z[i][j] for i in range(len(meals))) == num_meals)

    # HARD CONSTRAINT 2: Number of times a food is used
    for i, (meal, (low, high)) in enumerate(zip(meals, meals_limits)):
        times_used = sum(z[i])
        assert len(z[i]) == num_days

        # Add lower limit
        if low is not None:
            if low > num_days:
                msg = "Lower limit on '{}' is {}, but there are {} days."
                raise RuntimeError(msg.format(meal.name, low, num_days))
            solver.Add(times_used >= low)

        # Add upper limit
        if high is not None:
            solver.Add(times_used <= high)

    # =============================================================================
    #     SOLVE THE OPTIMIZATION PROBLEM
    # =============================================================================

    # Minimize the deviation from the goal
    solver.Minimize(objective_function)
    result_status = solver.Solve()
    if result_status == solver.INFEASIBLE:
        if first_call:
            params["first_call"] = False

            return optimize_mealplan(
                meals=meals,
                dietary_constraints=dietary_constraints,
                meals_limits=meals_limits,
                num_days=num_days,
                num_meals=num_meals,
                time_limit_secs=time_limit_secs,
                epsilon=epsilon,
                weight_price=weight_price,
                weight_nutrients=weight_nutrients,
                weight_range=weight_range,
                params=params,
            )

        else:
            raise RuntimeError("Infeasible problem.")

    assert solver.VerifySolution(1e-7, True)

    # =============================================================================
    #     POSTPROCESS THE SOLUTION AND RETURN
    # =============================================================================

    # Parse the variables and get the solution values
    for j in range(num_days):
        for i, meal in enumerate(meals):
            x[i][j] = x[i][j].solution_value()
            z[i][j] = z[i][j].solution_value()

            # Food is chosen
            if z[i][j] > 0.5:
                # If the food is chosen, x_ij is no smaller than epsilon
                x[i][j] = max(x[i][j], EPSILON)
            else:
                x[i][j] = 0

    # Compute the total price
    total_price = 0
    for j in range(num_days):
        daily_price = sum(x[i][j] * meal.price for i, meal in enumerate(meals))
        total_price += daily_price

    return (
        x,
        {
            "obj_func_value": round(solver.Objective().Value(), 6),
            "wall_time": round(solver.wall_time() / 1000, 3),
            "iterations": solver.iterations(),
            "total_price": round(total_price, 1),
        },
    )


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import statistics
import warnings
import numbers
import time
import typing
import inspect
from os import path

from jinja2 import Environment, FileSystemLoader

from streprogen.day import Day
from streprogen.exercises import DynamicExercise, StaticExercise
from streprogen.modeling import (
    progression_diffeq,
    reps_to_intensity,
)
from streprogen.optimization import RepSchemeOptimizer, RepSchemeGenerator
from streprogen.utils import (
    chunker,
    escape_string,
    prioritized_not_None,
    round_to_nearest,
)


class Program(object):
    """The program class is a container for days and exercises, along with
    the methods and functions used to create training programs."""

    REP_SET_SEP = " x "
    TEMPLATE_DIR = path.join(path.dirname(__file__), "templates")
    TEMPLATE_NAMES = {extension: "program_template." + extension for extension in ["html", "txt", "tex"]}

    # Default functions
    _default_rep_scaler_func = staticmethod(
        functools.partial(
            progression_diffeq,
            start_weight=1 + 0.2,
            final_weight=1 - 0.2,
            start_week=1,
            k=0,
        )
    )
    _default_intensity_scaler_func = staticmethod(
        functools.partial(
            progression_diffeq,
            start_weight=1 - 0.05,
            final_weight=1 + 0.05,
            start_week=1,
            k=0,
        )
    )

    _default_reps_to_intensity_func = staticmethod(reps_to_intensity)
    _default_progression_func = staticmethod(functools.partial(progression_diffeq, k=1))

    def __init__(
        self,
        name: str = "Untitled",
        duration: int = 8,
        reps_per_exercise: int = 25,
        min_reps: int = 3,
        max_reps: int = 8,
        rep_scaler_func: typing.Callable[[int], float] = None,
        intensity: float = 83,
        intensity_scaler_func: typing.Callable[[int], float] = None,
        units: str = "kg",
        round_to: typing.Union[float, typing.Callable] = 2.5,
        percent_inc_per_week: float = 1.5,
        progression_func: typing.Callable = None,
        reps_to_intensity_func: typing.Callable[[int], float] = None,
        verbose: bool = False,
    ):
        """Initialize a new program.


        Parameters
        ----------
        name
            The name of the training program, e.g. 'TommyAugust2017'.

        duration
            The duration of the training program in weeks, e.g. 8.

        reps_per_exercise
            The baseline number of repetitions per dynamic exercise.
            Typically a value in the range [15, 30].

        min_reps
            The minimum number of repetitions for the exercises, e.g. 3.
            This value can be set globally for the program, or for a specific
            dynamic exercise. If set at the dynamic exercise level, it will
            override the global program value.

        max_reps
            The maximum number of repetitions for the exercises, e.g. 8.
            This value can be set globally for the program, or for a specific
            dynamic exercise. If set at the dynamic exercise level, it will
            override the global program value.

        rep_scaler_func
            A function mapping from a week in the range [1, `duration`] to a
            scaling value (factor). The scaling value will be multiplied with
            the `reps_per_exercise` parameter for that week. Should typically
            return factors between 0.7 and 1.3.
            Alternatively, a list of length `duration` may be passed.

        intensity
            The baseline intensity for each dynamic exercise. The intensity
            of an exercise for a given week is how heavy the average
            repetition is compared to the expected 1RM (max weight one can
            lift) for that given week. Typically a value around 80.

        intensity_scaler_func
            A function mapping from a week in the range [1, `duration`] to a
            scaling value (factor). The scaling value will be multiplied with
            the `intensity` parameter for that week.
            Should typically return factors between 0.9 and 1.1.
            Alternatively, a list of length `duration` may be passed.

        units
            The units used for exporting and printing the program, e.g. 'kg'.

        round_to
            Round the dynamic exercise to the nearest multiple of this
            parameter. Typically 2.5, 5 or 10. A callable can alternatively be passed.
            This value can be set globally for the program, or for a specific
            dynamic exercise. If set at the dynamic exercise level, it will
            override the global program value.

        percent_inc_per_week
            If `final_weight` is not set, this value will be used. Percentage
            increase per week can be set globally for the program, or for each
            dynamic exercise. If set at the dynamic exercise level, it will
            override the global program value. The increase is  additive, not
            multipliactive. For instance, if the increase is set to
            `percent_inc_per_week=2`, then after 2 weeks the increase is 4,
            not (1.02 * 1.02 - 1) * 100 = 4.04. The `final_weight` parameter
            must be set to `None` for this parameter to have effect.

        progression_func
            The function used to model overall 1RM progression in the
            training program. The function must have a signature like:
                func(week, start_weight, final_weight, start_week, end_week)

        reps_to_intensity_func
            The function used to model the relationship between repetitions
            and intensity. Maps from a repetition to an intensity in the range 0-100.

        verbose
            If True, information will be outputted as the program is created.


        Returns
        -------
        Program
            A Program instance.


        Examples
        -------
        >>> program = Program('My training program')
        >>> program._rendered
        False

        """
        self.name = escape_string(name)

        assert isinstance(duration, numbers.Integral) and duration > 1
        self.duration = duration

        assert isinstance(reps_per_exercise, numbers.Integral) and reps_per_exercise > 0
        self.reps_per_exercise = reps_per_exercise

        assert isinstance(min_reps, numbers.Integral) and min_reps > 0
        assert isinstance(max_reps, numbers.Integral) and max_reps > 0
        self.min_reps = min_reps
        self.max_reps = max_reps
        if self.min_reps and self.max_reps:
            if self.min_reps > self.max_reps:
                raise ValueError("'min_reps' larger than 'max_reps'")

        assert isinstance(intensity, numbers.Number) and intensity > 0
        self.intensity = intensity

        assert isinstance(units, str)
        self.units = units
        self.round_to = round_to

        # Create a callable
        if callable(round_to):
            self.round = round_to
        else:
            self.round = functools.partial(round_to_nearest, nearest=round_to)

        self.verbose = verbose

        # ------ REP SCALERS ------
        # Set functions to user supplied, or defaults if None was passed
        user, default = (
            rep_scaler_func,
            functools.partial(self._default_rep_scaler_func, final_week=self.duration),
        )
        rep_scaler_func = prioritized_not_None(user, default)
        if callable(rep_scaler_func):
            self.rep_scalers = [rep_scaler_func(w + 1) for w in range(self.duration)]
            self.rep_scaler_func = rep_scaler_func
        else:
            self.rep_scalers = list(rep_scaler_func)
        assert isinstance(self.rep_scalers, list)

        # ------ INTENSITY SCALERS------
        user, default = (
            intensity_scaler_func,
            functools.partial(self._default_intensity_scaler_func, final_week=self.duration),
        )
        intensity_scaler_func = prioritized_not_None(user, default)
        if callable(intensity_scaler_func):
            self.intensity_scalers = [intensity_scaler_func(w + 1) for w in range(self.duration)]
            self.intensity_scaler_func = intensity_scaler_func
        else:
            self.intensity_scalers = list(intensity_scaler_func)
        assert isinstance(self.intensity_scalers, list)

        user, default = progression_func, self._default_progression_func
        self.progression_func = prioritized_not_None(user, default)
        assert callable(self.progression_func)

        user, default = reps_to_intensity_func, self._default_reps_to_intensity_func
        self.reps_to_intensity_func = prioritized_not_None(user, default)
        assert callable(self.reps_to_intensity_func)

        # Setup variables that the user has no control over
        self.days = []
        self.active_day = None  # Used for Program.Day context manager API
        self._rendered = False
        self._set_jinja2_enviroment()

        assert isinstance(percent_inc_per_week, numbers.Number)
        self.percent_inc_per_week = percent_inc_per_week

        # TODO: make explicit
        self.optimizer = RepSchemeOptimizer()

    def set_optimization_params(self, reps_slack=None, max_diff=None, max_unique=None):
        """Set default parameters.

        Passing None for a paramter will get use the defaults in RepSchemeGenerator.

        Parameters
        ----------
        reps_slack : int, optional
            Maximum deviation from the repetition goal.
        max_diff : int, optional
            Maximum difference between two consecutive sets.
        max_unique : int, optional
            Maximum unique sets in the solution.

        """

        # Get the default parameters
        signature = inspect.signature(RepSchemeGenerator.__init__)
        defaults = {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}

        # Use defaults if None is passed
        reps_slack = defaults["reps_slack"] if reps_slack is None else reps_slack
        max_diff = defaults["reps_slack"] if max_diff is None else max_diff
        max_unique = defaults["reps_slack"] if max_unique is None else max_unique

        self.optimizer = RepSchemeOptimizer(
            RepSchemeGenerator(reps_slack=reps_slack, max_diff=max_diff, max_unique=max_unique)
        )

    def serialize(self) -> dict:
        """Export the object to a dictionary."""
        data = {
            "name": self.name,
            "duration": self.duration,
            "min_reps": self.min_reps,
            "max_reps": self.max_reps,
            "reps_per_exercise": self.reps_per_exercise,
            "rep_scaler_func": self.rep_scalers,
            "intensity": self.intensity,
            "intensity_scaler_func": self.intensity_scalers,
            "units": self.units,
            "round_to": self.round_to,
            "percent_inc_per_week": self.percent_inc_per_week,
            "verbose": self.verbose,
        }

        data["days"] = [day.serialize() for day in self.days]
        return data

    @classmethod
    def deserialize(cls, data: dict):
        """Create a new object from a dictionary."""
        days = data.pop("days")
        new_program = cls(**data)

        for day_data in days:
            new_program.add_days(Day.deserialize(day_data))

        return new_program

    def Day(self, name: str = None):
        day = Day(name=name)
        day.program = self
        return day

    def DynamicExercise(
        self,
        name,
        start_weight=None,
        final_weight=None,
        min_reps=None,
        max_reps=None,
        percent_inc_per_week=None,
        reps=None,
        intensity=None,
        round_to=None,
        shift=0,
    ):
        ex = DynamicExercise(
            name,
            start_weight,
            final_weight,
            min_reps,
            max_reps,
            percent_inc_per_week,
            reps,
            intensity,
            round_to,
            shift,
        )
        self.active_day.exercises.append(ex)
        ex.day = self.active_day
        return ex

    def StaticExercise(self, name, sets_reps="4 x 10"):
        ex = StaticExercise(name, sets_reps)
        self.active_day.exercises.append(ex)
        return ex

    def _validate(self):
        """
        The purpose of this method is to verify that the user has set sensible
        values for the training program before rendering. The user will still
        be able to render, but error messages will be printed. This method:

            * Validates that the average intensity is in the range [65, 85].
            * Validates that the number of repetitions is in the range [15, 45].
            * Validates that 'reps_to_intensity_func' maps to [0, 100].
            * Validates that 'reps_to_intensity_func' is a decreasing function.
            * Validates that the exercises do not grow too much.

        Apart from these sanity checks, the user is on their own.
        """

        # Warn if many days have the same name
        if len(self.days) > len(set(day.name for day in self.days)):
            warnings.warn("\nWARNING: Some day names are duplicates.")

        weeks = list(range(self.duration))

        # Validate the intensity
        intensities = [self.intensity_scalers[w] * self.intensity for w in weeks]

        if max(intensities) > 90:
            warnings.warn("\nWARNING: At least one week has intensity > 90.")

        if min(intensities) < 65:
            warnings.warn("\nWARNING: At least one week has intensity < 65.")

        # Validate the repetitions
        repetitions = [self.rep_scalers[w] * self.reps_per_exercise for w in weeks]
        if max(repetitions) > 45:
            warnings.warn("\nWARNING: At least one week has repetitions > 45.")

        if min(repetitions) < 15:
            warnings.warn("\nWARNING: At least one week has repetitions < 15.")

        # Validate the 'reps_to_intensity_func'
        for x1, x2 in zip(range(1, 20), range(2, 21)):
            y1 = self.reps_to_intensity_func(x1)
            y2 = self.reps_to_intensity_func(x2)
            if y1 < y2:
                warnings.warn("\n'reps_to_intensity_func' is not decreasing.")

        if any(self.reps_to_intensity_func(x) > 100 for x in range(1, 20)):
            warnings.warn("\n'reps_to_intensity_func' maps to > 100.")

        if any(self.reps_to_intensity_func(x) < 0 for x in range(1, 20)):
            warnings.warn("\n'reps_to_intensity_func' maps to < 0.")

        # Validate the exercises
        for exercise in self._yield_exercises():
            if isinstance(exercise, StaticExercise):
                continue

            _, _, percent_inc_per_week = exercise._progress_information()
            if percent_inc_per_week > 4:
                msg = f'\n"{exercise.name}" grows with {percent_inc_per_week}% each week.'
                warnings.warn(msg)

    def add_days(self, *days):
        """Add one or several days to the program.

        Parameters
        ----------
        *days
            Iterable containing
            :py:class:`streprogen.Day` instances.

        Examples
        -------
        >>> program = Program('My training program')
        >>> day1, day2 = Day(), Day()
        >>> program.add_days(day1, day2)
        >>> program.add_days(day1)
        """
        for day in days:
            day.program = self
            self.days.append(day)

    def _render_dynamic(self, dynamic_exercise, desired_reps, desired_intensity, validate) -> dict:
        """
        Render a single dynamic exercise.
        This is done for every exercise for every week.
        """

        min_reps = dynamic_exercise.min_reps
        max_reps = dynamic_exercise.max_reps

        # Use tuples as inputs to the optimizer can cache the arguments
        sets = tuple(range(min_reps, max_reps + 1))
        intensities = tuple(map(self.reps_to_intensity_func, sets))

        reps = self.optimizer(
            sets=sets, intensities=intensities, reps_goal=desired_reps, intensity_goal=desired_intensity
        )

        intensities = list(map(self.reps_to_intensity_func, reps))

        # If repetitions are too high, a low average intensity cannot be attained
        int_highest = self.reps_to_intensity_func(min_reps)
        int_lowest = self.reps_to_intensity_func(max_reps)

        if (not (int_lowest - 0.1 <= desired_intensity <= int_highest + 0.1)) and validate:
            msg = """WARNING: The exercise '{}' is restricted to repetitions in the range [{}, {}].
This maps to intensities in the range [{}, {}], but the goal average intensity is {},
which is not achievable with this rep range.
SOLUTION: Either (1) change the repetition range, (2) change the desired intensity
or (3) ignore this message. The software will do it's best to remedy this.
""".format(
                dynamic_exercise.name,
                min_reps,
                max_reps,
                round(int_lowest, 1),
                round(int_highest, 1),
                round(desired_intensity, 1),
            )
            warnings.warn(msg)

        return {"reps": reps, "intensities": intensities}

    def _initialize_render_dictionary(self):
        """Initialize a dictionary for rendered values.

        Examples
        -------
        >>> program = Program('My training program')
        >>> program._initialize_render_dictionary()
        >>> program._rendered is False
        False
        """

        self._rendered = dict()

        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            self._rendered[week] = dict()

            # Iterate over all days
            for day in self.days:
                self._rendered[week][day] = dict()

                # Iterate over all main exercises
                for exercise in day.exercises:
                    if isinstance(exercise, DynamicExercise):
                        self._rendered[week][day][exercise] = dict()

    def _yield_week_day_dynamic(self):
        """A helper function to reduce the number of nested loops.

        Yields
        -------
        tuple
            A tuple with (week, day, dynamic_ex).

        """
        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            # Iterate over all days
            for day in self.days:
                # Iterate over all main exercises
                for exercise in day.exercises:
                    if not isinstance(exercise, DynamicExercise):
                        continue
                    yield (week, day, exercise)

    def _yield_exercises(self):
        """A helper function to reduce the number of nested loops.


        Yields
        -------
        (dynamic_ex) or (static_ex)
            Yields the exercises in the program.

        """
        for day in self.days:
            for exercise in day.exercises:
                yield exercise

    def render(self, validate=True):
        """Render the training program to perform the calculations.
        The program can be rendered several times to produce new
        information given the same input parameters.

        Parameters
        ----------
        validate
            Boolean that indicates whether or not to run a validation
            heurestic on the program before rendering. The validation
            will warn the user if inputs seem unreasonable.

        """
        start_time = time.time()

        # Check that exercise names are unique within each day
        for day in self.days:
            seen_names = set()
            for exercise in day.exercises:
                if exercise.name in seen_names:
                    raise ValueError(f"Exercise name not unique: {exercise.name}")
                else:
                    seen_names.add(exercise.name)

        # --------------------------------
        # Prepare for rendering the dynamic exercises
        # --------------------------------

        # Initialize the structure of the _rendered dictionary
        self._initialize_render_dictionary()

        # Set the day names
        for i, day in enumerate(self.days):
            day.name = prioritized_not_None(day.name, "Day {}".format(i + 1))

        # Validate the program if the user wishes to validate
        if validate:
            self._validate()

        # --------------------------------
        # Render the dynamic exercises
        # --------------------------------

        for week, day, dyn_ex in self._yield_week_day_dynamic():
            # Set min and max reps from program, if not set on exercise
            min_reps = dyn_ex.min_reps
            max_reps = dyn_ex.max_reps

            if min_reps > max_reps:
                msg = "'min_reps' larger than 'max_reps' for exercise '{}'."
                raise ValueError(msg.format(dyn_ex.name))

            # Use the local rounding function if available,
            # if not use the global rounding function
            round_func = prioritized_not_None(dyn_ex.round, self.round)

            # The desired repetitions to work up to
            total_reps = prioritized_not_None(dyn_ex.reps, self.reps_per_exercise)

            # If the index is not valid (due to shifting), use function
            index_to_lookup = week - 1 + dyn_ex.shift
            if 0 <= index_to_lookup < self.duration:
                desired_reps = round(total_reps * self.rep_scalers[index_to_lookup])
            else:
                if hasattr(self, "rep_scaler_func"):
                    desired_reps = round(total_reps * self.rep_scaler_func(week + dyn_ex.shift))
                else:
                    raise TypeError("Using `shift` requires `rep_scaler_func` to be a function, not a list.")

            self._rendered[week][day][dyn_ex]["desired_reps"] = int(desired_reps)

            # The desired average intensity
            intensity_unscaled = prioritized_not_None(dyn_ex.intensity, self.intensity)
            if 0 <= index_to_lookup < self.duration:
                scale_factor = self.intensity_scalers[index_to_lookup]
            else:
                if hasattr(self, "intensity_scaler_func"):
                    scale_factor = self.intensity_scaler_func(week + dyn_ex.shift)
                else:
                    raise TypeError("Using `shift` requires `intensity_scaler_func` to be a function, not a list.")

            desired_intensity = intensity_unscaled * scale_factor
            self._rendered[week][day][dyn_ex]["desired_intensity"] = desired_intensity

            # A dictionary is returned with keys 'reps' and 'intensities'
            render_args = dyn_ex, desired_reps, desired_intensity, validate
            out = self._render_dynamic(*render_args)

            # Get increase from program if not available on the exercise
            inc_week = prioritized_not_None(dyn_ex.percent_inc_per_week, self.percent_inc_per_week)

            # Compute the progress
            (start_w, final_w, inc_week) = dyn_ex._progress_information()

            weight = self.progression_func(week + dyn_ex.shift, start_w, final_w, 1, self.duration)

            # Test that the weight is not too far from min and max
            upper_threshold = max(start_w, final_w) + abs(start_w - final_w)
            lower_threshold = min(start_w, final_w) - 2 * abs(start_w - final_w)
            if not (lower_threshold <= weight <= upper_threshold):
                msg = f"\nWARNING: Weight for '{dyn_ex.name}' was {round(weight, 2)} in week {week}. "
                msg += f"This is far from start and final weights. Start weight is {start_w}. "
                msg += f"Final weight is {final_w}."
                warnings.warn(msg)

            # Define a function to prettify the weights
            def pretty_weight(weight, i, round_function):
                weight = round_function(weight * i / 100)
                if weight % 1 == 0:
                    return int(weight)
                return weight

            # Create pretty strings
            tuples_gen = zip(out["intensities"], out["reps"])
            pretty_gen = ((str(r), str(pretty_weight(weight, i, round_func)) + self.units) for (i, r) in tuples_gen)
            out["strings"] = list(self.REP_SET_SEP.join(list(k)) for k in pretty_gen)
            out["1RM_this_week"] = round(weight, 2)
            out["weights"] = [pretty_weight(weight, i, round_func) for i in out["intensities"]]

            # Update with the ['intensities', 'reps', 'strings', ...] keys
            self._rendered[week][day][dyn_ex].update(out)

        if self.verbose:
            delta_time = round(time.time() - start_time, 3)
            print(f"Rendered program in {delta_time} seconds.")

    def to_dict(self):
        """Write the rendered program information to a dictionary."""
        if not hasattr(self, "_rendered"):
            raise ValueError("Render the program by calling .render() first.")

        output_dictionary = {"program": self.serialize()}
        output_dictionary["rendered"] = []

        if not hasattr(self, "_rendered"):
            return output_dictionary

        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            output_week = []

            # Iterate over all days
            for day in self.days:
                output_day = {"name": day.name, "exercises": []}

                for exercise in day.exercises:
                    if isinstance(exercise, DynamicExercise):
                        out = exercise.serialize()
                        out.update(self._rendered[week][day][exercise])
                        output_day["exercises"].append(out)
                    elif isinstance(exercise, StaticExercise):
                        output_day["exercises"].append(exercise.serialize())
                    else:
                        raise TypeError(f"Expected exercise, got object: {exercise}")

                # Add daily
                output_week.append(output_day)

            # Add weekly dict
            output_dictionary["rendered"].append(output_week)

        return output_dictionary

    def _set_jinja2_enviroment(self):
        """
        Set up the jinja2 environment.
        """

        template_loader = FileSystemLoader(searchpath=self.TEMPLATE_DIR)

        env = Environment(loader=template_loader, trim_blocks=True, lstrip_blocks=True)
        env.globals.update(chunker=chunker, enumerate=enumerate, str=str)

        # Add filters to the environment
        round2digits = functools.partial(round_to_nearest, nearest=0.1)
        env.filters["round2digits"] = round2digits
        env.filters["mean"] = statistics.mean

        def is_static_exercise(arg):
            return isinstance(arg, StaticExercise)

        def is_dynamic_exercise(arg):
            return isinstance(arg, DynamicExercise)

        env.filters["is_static_exercise"] = is_static_exercise
        env.filters["is_dynamic_exercise"] = is_dynamic_exercise

        self.jinja2_environment = env

    def to_html(self, table_width=5):
        """Write the program information to HTML code, which can be saved,
        printed and brought to the gym.

        Parameters
        ----------
        table_width
            The table with of the HTML code.

        Returns
        -------
        string
            HTML code.
        """

        env = self.jinja2_environment
        template = env.get_template(self.TEMPLATE_NAMES["html"])
        return template.render(program=self, table_width=table_width)

    def to_txt(self, verbose=False):
        """Write the program information to text,
        which can be printed in a terminal.

        Parameters
        ----------
        verbose
            If True, more information is shown.

        Returns
        -------
        string
            Program as text.
        """
        # Get information related to formatting
        exercises = list(self._yield_exercises())
        max_ex_name = 0
        if exercises:
            max_ex_name = max(len(ex.name) for ex in exercises)

        # If rendered, find the length of the longest '6 x 75kg'-type string
        max_ex_scheme = 0
        if self._rendered:
            for week, day, dynamic_ex in self._yield_week_day_dynamic():
                lengths = [len(s) for s in self._rendered[week][day][dynamic_ex]["strings"]]
                max_ex_scheme = max(max_ex_scheme, max(lengths))

        env = self.jinja2_environment
        template = env.get_template(self.TEMPLATE_NAMES["txt"])
        return template.render(
            program=self,
            max_ex_name=max_ex_name,
            max_ex_scheme=max_ex_scheme,
            verbose=verbose,
        )

    def to_tex(self, text_size="large", table_width=5, clear_pages=False):
        r"""
        Write the program information to a .tex file, which can be
        rendered to .pdf running pdflatex. The program can then be
        printed and brought to the gym.

        Parameters
        ----------
        text_size
            The tex text size, e.g. '\small', 'normalsize', 'large', 'Large'
            or 'LARGE'.

        table_width
            The table width of the .tex code.

        clear_pages
            If True, the page will be cleared after each week is printed.

        Returns
        -------
        string
            Program as tex.
        """

        # If rendered, find the length of the longest '6 x 75kg'-type string
        max_ex_scheme = 0
        if self._rendered:
            for week, day, dynamic_ex in self._yield_week_day_dynamic():
                lengths = [len(s) for s in self._rendered[week][day][dynamic_ex]["strings"]]
                max_ex_scheme = max(max_ex_scheme, max(lengths))

        env = self.jinja2_environment
        template = env.get_template(self.TEMPLATE_NAMES["tex"])

        return template.render(
            program=self,
            text_size=text_size,
            table_width=table_width,
            clear_pages=clear_pages,
        )

    def __str__(self):
        """
        String formatting for readable human output.
        """
        return self.to_txt()


# Patch up the docs
Program.Day.__doc__ = Day.__doc__ + "\nSee streprogen.Day for accurate signature."
Program.DynamicExercise.__doc__ = DynamicExercise.__doc__ + "\nSee streprogen.DynamicExercise for accurate signature."
Program.StaticExercise.__doc__ = StaticExercise.__doc__ + "\nSee streprogen.StaticExercise for accurate signature."

if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])

if __name__ == "__main__":
    from streprogen import Program

    def rep_scaler_func(week, *args):
        return 1

    def intensity_scaler_func(week, *args):
        return 1

    program = Program(
        # The name of the training program
        name="Beginner 5x5",
        # The duration of the training program in weeks.
        duration=4,
        # The baseline number of repetitions per dynamic exercise.
        # reps_per_exercise=25,
        intensity=reps_to_intensity(5),
        # Units for the weights, typically 'kg', 'lbs' or '' (empty)
        units="kg",
        # What the weights are rounded to.
        round_to=2.5,
        rep_scaler_func=[1, 1, 1, 1],
        intensity_scaler_func=intensity_scaler_func,
    )

    with program.Day("A"):
        program.StaticExercise("Biceps", "4 x 10")
        program.DynamicExercise(name="Squat", start_weight=100, min_reps=5, max_reps=5)

    program.render()

    # pprint(program.to_dict())

    print(program)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import warnings
import inspect

from streprogen.utils import compose, escape_string, round_to_nearest, prioritized_not_None


class DynamicExercise(object):
    """Class for dynamic exercises."""

    def __init__(
        self,
        name,
        start_weight=None,
        final_weight=None,
        min_reps=None,
        max_reps=None,
        percent_inc_per_week=None,
        reps=None,
        intensity=None,
        round_to=None,
        shift=0,
    ):
        """Initialize a new dynamic exercise. A dynamic exercise is rendered by
        the program, and the set/rep scheme will vary from week to week.

        Parameters
        ----------
        name
            The name of the exercise, e.g. 'Squats'.

        start_weight
            Maximum weight you can lift at the start of the program, e.g. 80.

        final_weight
            The goal weight to work towards during the program. This should be
            set in relation to the duration of the training program, e.g. 90.
            If set, this overrides the optional `percent_inc_per_week` parameter.

        min_reps
            The minimum number of repetitions for this exercise, e.g. 3.

        max_reps
            The maximum number of repetitions for this exercise, e.g. 8.

        percent_inc_per_week
            If `final_weight` is not set, this value will be used. The increase is
            additive, not multipliactive. For instance, if the increase is set to
            `percent_inc_per_week=2`, then after 2 weeks the increase is 4, not
            (1.02 * 1.02 - 1) * 100 = 4.04. The `final_weight` parameter must be
            set to `None` for this parameter to have effect.

        reps
            The number of baseline repetitions for this exercise. If this
            parameter is set, it will override the global 'reps_per_exercise'
            parameter for the training program. The repetitions will still
            be scaled by the `rep_scaler_func` parameter in the training program.

        intensity
            The average intensity for this exercise. If set, this will
            override the `intensity` parameter in the training program.
            The intensity will still be scaled by the `intensity_scaler_func`
            parameter.

        round_to
            Round the output to the closest multiple of this number, e.g. 2.5.
            Alternatively, a callable can be passed. The callable must take
            a number as input and return a rounded number as output.

        shift
            Shift evaluation of repetitions, intensity and progress `shift`
            weeks ahead in time. An exercise shifted by 1 will have its reps,
            intensity and strength evalated at week i + 1 instead of in week i.


        Examples
        -------
        >>> bench = DynamicExercise('Bench press', 100, 120, 3, 8)
        >>> bench2 = DynamicExercise('Bench press', 100, 120, 3, 8)
        >>> bench == bench2
        True


        """
        self.name = escape_string(name)
        self.start_weight = start_weight
        self.final_weight = final_weight
        self._min_reps = min_reps
        self._max_reps = max_reps
        self.percent_inc_per_week = percent_inc_per_week
        self.reps = reps
        self.intensity = intensity
        self.day = None

        var_names = ["start_weight", "final_weight", "percent_inc_per_week"]
        num_specified = sum(1 if (getattr(self, var) is not None) else 0 for var in var_names)
        if num_specified == 3:
            raise ValueError(f"At most 2 out of 3 variables may be set: {var_names}")

        self.round_to = round_to

        assert isinstance(shift, int), "'shift' must be an integer"
        # assert shift >= 0, "'shift' must be non-negative"
        self.shift = shift

        # Use program value instead of exercise value
        if round_to is None:
            self.round = None

        # A callable
        elif callable(round_to):
            self.round = round_to
        else:
            self.round = functools.partial(round_to_nearest, nearest=round_to)

        if self.final_weight and self.start_weight:
            if self.start_weight > self.final_weight:
                msg = "'start_weight' larger than 'final_weight' for exercise '{}'."
                warnings.warn(msg.format(self.name))

        if self.min_reps and self.max_reps:
            if self.min_reps > self.max_reps:
                msg = "'min_reps' larger than 'max_reps' for exercise '{}'."
                raise ValueError(msg.format(self.name))

    def _simple_attributes(self):
        """Yield all simple parameters (ints, strings, etc)."""
        attributes = list(dir(self))  # round is a function
        attributes.remove("round")
        attributes.remove("day")
        attributes.remove("shift")

        for attr_name in attributes:
            # Skip private members
            if attr_name.startswith("_"):
                continue

            # Skip methods
            if inspect.ismethod(getattr(self, attr_name)):
                continue

            if getattr(self, attr_name):
                yield attr_name

    def serialize(self):
        """Export the object to a dictionary.

        Examples
        --------
        >>> bench = DynamicExercise('Bench press', start_weight=100)
        >>> bench.serialize() == {'start_weight': 100, 'name': 'Bench press'}
        True

        """
        result = {
            "name": self.name,
            "start_weight": self.start_weight,
            "final_weight": self.final_weight,
            "min_reps": self._min_reps,
            "max_reps": self._max_reps,
            "percent_inc_per_week": self.percent_inc_per_week,
            "reps": self.reps,
            "intensity": self.intensity,
            "round_to": self.round_to,
        }

        return {k: v for (k, v) in result.items() if v}

    def _progress_information(self):
        """Return a tuple (start_weight, final_weight, percent_inc_per_week).

        Can only be inferred in the context of a Program argument.

        """

        if self.day is None:
            raise Exception("Exercise {self.name} must be attached to a Day.")

        if self.day.program is None:
            raise Exception("Day {self.day.name} must be attached to a Program.")

        program = self.day.program

        # Get increase per week
        inc_week = prioritized_not_None(self.percent_inc_per_week, program.percent_inc_per_week)

        # Case 1: Start weight and final weight is given
        if (self.start_weight is not None) and (self.final_weight is not None):
            start_w, final_w = self.start_weight, self.final_weight
            inc_week = ((final_w / start_w) - 1) / program.duration * 100
            answer = (start_w, final_w, inc_week)

        # Case 2: Start weight and increase is given
        elif (self.start_weight is not None) and (inc_week is not None):
            factor = 1 + (inc_week / 100) * program.duration
            start_w = self.start_weight
            final_w = self.start_weight * factor
            answer = (start_w, final_w, inc_week)

        # Case 3: Final weight and increase is given
        elif (self.final_weight is not None) and (inc_week is not None):
            factor = 1 + (inc_week / 100) * program.duration
            start_w = self.final_weight / factor
            final_w = self.final_weight
            answer = (start_w, final_w, inc_week)

        else:
            raise Exception(f"Exercise {self} is overspecified.")

        rounder = functools.partial(round_to_nearest, nearest=0.01)

        return tuple(map(rounder, answer))

    @property
    def min_reps(self):
        """Return min reps. If a Program attribute it set and the exercise
        attribute is None, use the program attribute."""
        if (self.day is not None) and (self.day.program is not None):
            program = self.day.program
        else:
            return self._min_reps

        return prioritized_not_None(self._min_reps, program.min_reps)

    @min_reps.setter
    def min_reps(self):
        return self._min_reps

    @property
    def max_reps(self):
        """Return max reps. If a Program attribute it set and the exercise
        attribute is None, use the program attribute."""
        if (self.day is not None) and (self.day.program is not None):
            program = self.day.program
        else:
            return self._max_reps

        return prioritized_not_None(self._max_reps, program.max_reps)

    @max_reps.setter
    def max_reps(self):
        return self._max_reps

    def __repr__(self):
        """Representation."""
        return str(self)

    def __str__(self):
        """Human readable output.

        Examples
        --------
        >>> ex = DynamicExercise(name="Bench", start_weight=100)
        >>> str(ex)
        "DynamicExercise(name='Bench', start_weight=100)"
        """

        attr_names = self._simple_attributes()
        arg_str = ", ".join(["{}={}".format(attr, repr(getattr(self, attr))) for attr in attr_names])
        return "{}({})".format(type(self).__name__, arg_str)

    def __eq__(self, other):
        return all(getattr(self, attr) == getattr(other, attr) for attr in self._simple_attributes())

    def __hash__(self):
        attrs = ("name",)
        return hash(tuple(getattr(self, attr) for attr in attrs))

    @classmethod
    def deserialize(cls, data):
        """Create a new object from a dictionary."""
        return cls(**data)


class StaticExercise(object):
    """
    Class for static exercises.
    """

    def __init__(self, name, sets_reps="4 x 10"):
        """Initialize a new static exercise. A static exercise
        is simply a placeholder for some text.

        Parameters
        ----------
        name
            The name of the exercise, e.g. 'Curls'.

        sets_reps
            A static set/rep scheme, e.g. '4 x 10', or '10 minutes'.
            This paramter can also be a function of one parameter,
            the current week. The function must return a string
            for that specific week.


        Returns
        -------
        StaticExercise
            A StaticExercise object.


        Examples
        -------
        >>> curls = StaticExercise('Curls', '4 x 10')
        >>> stretching = StaticExercise('Stretching', '10 minutes')
        """
        self.name = escape_string(name)
        self.sets_reps = sets_reps
        if isinstance(sets_reps, str):
            self.sets_reps_func = self._function_from_string(sets_reps)
        else:
            self.sets_reps_func = sets_reps

        # Escape after function evaluation
        self.sets_reps_func = compose(self.sets_reps_func, escape_string)

    @staticmethod
    def _function_from_string(string):
        """
        Static method that takes a string and returns a function which returns
        the string.
        """

        def function(*args, **kwargs):
            return string

        return function

    def __repr__(self):
        """
        Representation.
        """
        return "{}({})".format(type(self).__name__, str(self.__dict__)[:60])

    def __eq__(self, other):
        return self.name == other.name and self.sets_reps == other.sets_reps

    def __str__(self):
        """
        String formatting for readable human output.
        """

        strvar = ["name", "sets_reps"]

        arg_str = ", ".join(["{}={}".format(k, self.__dict__[k]) for k in strvar if self.__dict__[k] is not None])

        return "{}({})".format(type(self).__name__, arg_str)

    def serialize(self):
        if callable(self.sets_reps):
            raise ValueError(f"Cannot serialize {repr(self)} because `sets_reps` is a function.")
        return {"name": self.name, "sets_reps": self.sets_reps}

    @classmethod
    def deserialize(cls, data):
        return cls(**data)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-vv", "--capture=sys", "-k "])
    bench = DynamicExercise("Bench", start_weight=100)
    print(bench.serialize())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from streprogen.optimization import RepSchemeOptimizer, RepSchemeGenerator
from streprogen.modeling import (
    reps_to_intensity,
    reps_to_intensity_relaxed,
    reps_to_intensity_tight,
)
import itertools
import pytest


@pytest.mark.parametrize(
    "reps_to_intensity_func, reps_goal, intensities_goal",
    list(
        itertools.product(
            [reps_to_intensity, reps_to_intensity_relaxed, reps_to_intensity_tight],
            list(range(10, 35 + 1, 2)),
            list(range(70, 95 + 1, 2)),
        )
    ),
)
def test_repscheme_optimizer(reps_to_intensity_func, reps_goal, intensities_goal):
    """For common settings, check that the optimization returns good results."""

    # Prepare data
    reps = tuple(range(1, 12 + 1))
    intensities = tuple(reps_to_intensity_func(r) for r in reps)

    # Default values make sense to use here
    optimizer = RepSchemeOptimizer()

    scheme = optimizer(sets=reps, intensities=intensities, reps_goal=reps_goal, intensity_goal=intensities_goal)
    assert scheme == sorted(scheme, reverse=True)
    assert scheme

    intensities = list(map(reps_to_intensity_func, scheme))

    reps = sum(scheme)
    intensity = sum(r * i for r, i in zip(scheme, intensities)) / reps

    assert abs(reps - reps_goal) <= 3
    assert abs(intensity - intensities_goal) <= 3.5


@pytest.mark.parametrize("reps_slack", [0, 1, 2, 3, 4])
def test_repscheme_generator_reps_slack(reps_slack):
    generator = RepSchemeGenerator(reps_slack=reps_slack)
    sets = [3, 4, 5, 6, 7, 8]
    for reps_goal in [15, 20, 25, 30]:
        for scheme in generator.generate(sets=sets, reps_goal=reps_goal):
            assert scheme
            assert abs(sum(scheme) - reps_goal) <= reps_slack


@pytest.mark.parametrize("max_unique", [1, 2, 3, 4])
def test_repscheme_generator_max_unique(max_unique):
    generator = RepSchemeGenerator(max_unique=max_unique)
    sets = [3, 4, 5, 6, 7, 8]
    for reps_goal in [15, 20, 25, 30]:
        for scheme in generator.generate(sets=sets, reps_goal=reps_goal):
            assert scheme
            assert len(set(scheme)) <= max_unique


@pytest.mark.parametrize("max_diff", [0, 1, 2, 3, 4])
def test_repscheme_generator_max_diff(max_diff):
    generator = RepSchemeGenerator(max_diff=max_diff)
    sets = [3, 4, 5, 6, 7, 8]
    for reps_goal in [15, 20, 25, 30]:
        for scheme in generator.generate(sets=sets, reps_goal=reps_goal):
            assert scheme
            if len(scheme) < 2:
                continue
            for s_j, s_i in zip(scheme[1:], scheme):
                assert s_j >= s_i
                assert s_j - s_i <= max_diff


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from streprogen.exercises import DynamicExercise, StaticExercise
from streprogen.utils import escape_string


class Day(object):
    """
    A day object is a container for exercises associated with the specific day.
    """

    def __init__(self, name=None, exercises=None):
        """Initialize a new day object.

        Parameters
        ----------
        name
            The name of the day, e.g. 'Day A'. If no name is given then the day
            will automatically be given a numeric name such as 'Day 1', 'Day 2', etc.

        exercises
            A list of exercises. Exercises can also be associated with a day using
            the 'add_exercises' method later on.


        Examples
        -------
        >>> monday = Day(name='Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> monday.add_exercises(curls)
        >>> curls in monday.exercises
        True
        """
        self.name = escape_string(name)
        self.exercises = []

        self.program = None

        if exercises is not None:
            self.add_exercises(*tuple(exercises))

    def __enter__(self):
        self.program.active_day = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.program.days.append(self)
        self.program.active_day = None

    def add_exercises(self, *exercises):
        """Add the exercises to the day. The method will automatically infer
        whether a static or dynamic exercise is passed to it.

        Parameters
        ----------
        *exercises
            An unpacked tuple of exercises.


        Examples
        -------
        >>> monday = Day(name='Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> pulldowns = StaticExercise('Pulldowns', '4 x 10')
        >>> monday.add_exercises(curls, pulldowns)
        >>> curls in monday.exercises
        True
        >>> pulldowns in monday.exercises
        True
        """
        for exercise in exercises:
            if isinstance(exercise, DynamicExercise):
                self.exercises.append(exercise)
                exercise.day = self

            if isinstance(exercise, StaticExercise):
                self.exercises.append(exercise)

    def __repr__(self):
        return "{}({})".format(type(self).__name__, str(self.__dict__)[:60])

    def __str__(self):
        """
        String formatting for readable human output.

        Examples
        -------
        >>> monday = Day(name = 'Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> pulldowns = StaticExercise('Pulldowns', '4 x 10')
        >>> monday.add_exercises(curls, pulldowns)
        >>> print(monday)
        Day(name = Monday, exercises = [Curls, Pulldowns])
        """
        exercises = ", ".join([d.name for d in self.exercises])
        if len(exercises) > 0:
            exercises = "exercises = [{}]".format(exercises)

        out_str = "{}({})".format(
            type(self).__name__,
            ", ".join([s for s in ["name = {}".format(self.name), exercises] if len(s) > 2]),
        )
        return out_str

    def serialize(self):
        """Export the object to a dictionary."""
        result = {"name": self.name}
        result["exercises"] = [ex.serialize() for ex in self.exercises]
        return result

    @classmethod
    def deserialize(cls, data):
        """Create a new object from a dictionary."""
        name = data["name"]
        exercises = [
            StaticExercise.deserialize(ex) if "sets_reps" in ex.keys() else DynamicExercise.deserialize(ex)
            for ex in data["exercises"]
        ]
        return cls(name, exercises)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import collections.abc
import functools
import math


def reps_to_intensity(reps, slope=-3.5, constant=97.5, quadratic=True):
    """A mapping from repetitions in range [1, 12] to intensities in range [0, 100].

    Parameters
    ----------
    reps
        The number of repetitions to map to the intensity range.
    slope
        Slope for the linear function.
    constant
        Constant for the linear function.
    quadratic
        If 'True', add a slight quadratic offset.


    Returns
    -------
    intensity
        An intensity value in the range from 0 to 100.


    Examples
    -------
    >>> reps_to_intensity(5, slope=-5, constant=100, quadratic=False)
    80
    >>> reps_to_intensity(8, slope=-5, constant=100, quadratic=True)
    67.45
    >>> reps_to_intensity(8, slope=-5, constant=100, quadratic=False)
    65
    """
    if isinstance(reps, collections.abc.Iterable):
        return list(reps_to_intensity(rep, slope, constant, quadratic) for rep in reps)

    intensity = constant + slope * (reps - 1)
    if quadratic:
        return intensity + 0.05 * (reps - 1) ** 2
    else:
        return intensity


def progression_sinusoidal(
    week,
    start_weight,
    final_weight,
    start_week,
    final_week,
    period=4,
    scale=0.025,
    offset=0,
    k=0,
    correct_boundaries=False,
):
    """A sinusoidal progression function going through the points
    ('start_week', 'start_weight') and ('final_week', 'final_weight'), evaluated
    in 'week'. This function calls a linear progression function
    and multiplies it by a sinusoid.

    Parameters
    ----------
    week
        The week to evaluate the linear function at.
    start_weight
        The weight at 'start_week'.
    final_weight
        The weight at 'final_week'.
    start_week
        The number of the first week, typically 1.
    final_week
        The number of the final week, e.g. 8.
    period
        The length of a period (cycle) in weeks.
    scale
        The scale (amplitude) of the sinusoidal term.
    offset
        The offset (shift) of the sinusoid.
    k
        Exponential growth. Higher results in more exponential growth.
    correct_boundaries
        If True, the function is adjusted to pass through the endpoints exactly.


    Returns
    -------
    weight
        The weight at 'week'.


    Examples
    -------
    >>> progression_sinusoidal(1, 123, 123, 1, 8, period=4)
    123.0
    >>> progression_sinusoidal(7, 123, 123, 1, 8, period=4)
    123.0
    """
    if isinstance(week, collections.abc.Iterable):
        return list(
            progression_sinusoidal(
                w, start_weight, final_weight, start_week, final_week, period, scale, offset, k, correct_boundaries
            )
            for w in week
        )

    # Get the base model
    base = progression_diffeq(week, start_weight, final_weight, start_week, final_week, k)
    if period <= 1:
        period = 1

    sine_argument = (week - offset - start_week) * (math.pi * 2) / period

    base_with_sinusoidal = base * (1 + scale * math.sin(sine_argument))

    if correct_boundaries:
        overshoot_start = start_weight - progression_sinusoidal(
            start_week,
            start_weight,
            final_weight,
            start_week,
            final_week,
            period,
            scale,
            offset,
            k,
            correct_boundaries=False,
        )

        undershoot_end = (
            progression_sinusoidal(
                final_week,
                start_weight,
                final_weight,
                start_week,
                final_week,
                period,
                scale,
                offset,
                k,
                correct_boundaries=False,
            )
            - final_weight
        )

        # Linear correction
        dy_dx = (overshoot_start + undershoot_end) / (final_week - start_week)
        correction = overshoot_start - dy_dx * (week - start_week)
        base_with_sinusoidal = base_with_sinusoidal + correction

    return base_with_sinusoidal


def progression_sawtooth(
    week,
    start_weight,
    final_weight,
    start_week,
    final_week,
    period=4,
    scale=0.025,
    offset=0,
    k=0,
    correct_boundaries=False,
):
    """A sawtooth progression function going through the points
    ('start_week', 'start_weight') and ('final_week', 'final_weight'), evaluated
    in 'week'. This function calls a linear progression function
    and multiplies it by a sawtooth.

    Parameters
    ----------
    week
        The week to evaluate the linear function at.
    start_weight
        The weight at 'start_week'.
    final_weight
        The weight at 'final_week'.
    start_week
        The number of the first week, typically 1.
    final_week
        The number of the final week, e.g. 8.
    period
        The length of a period (cycle) in weeks.
    scale
        The scale (amplitude) of the sawtooth term.
    offset
        The offset (shift) of the sawtooth.
    k
        Exponential growth. Higher results in more exponential growth.
    correct_boundaries
        If True, the function is adjusted to pass through the endpoints exactly.


    Returns
    -------
    weight
        The weight at 'week'.


    Examples
    -------
    >>> progression_sawtooth(1, 100, 100, 1, 8, period=8, scale=0.1)
    90.0
    >>> x = progression_sawtooth(8, 100, 100, 1, 8, period=8, scale=0.1)
    >>> int(x)
    110
    """
    if isinstance(week, collections.abc.Iterable):
        return list(
            progression_sawtooth(
                w, start_weight, final_weight, start_week, final_week, period, scale, offset, k, correct_boundaries
            )
            for w in week
        )

    # Get the base model
    base = progression_diffeq(week, start_weight, final_weight, start_week, final_week, k)
    if period <= 1:
        period = 1

    if period > 1:
        x = (week - offset - start_week) / period
        # https://en.wikipedia.org/wiki/Sawtooth_wave
        x = x - math.floor(x)

        # Due to the discrete nature of the sawtooth, the max is p-1 / p, not 1
        wave_amplitude = (period - 1) / period
        # Change the output to be in range: scale * [-1, 1]. The scaling ensures
        # that `scale` means the same thing in sinusoidal and triangle waveforms.
        saw = (2 / wave_amplitude) * x - 1
    else:
        saw = 0

    base_with_sinusoidal = base * (1 + scale * saw)

    if correct_boundaries:
        overshoot_start = start_weight - progression_sawtooth(
            start_week,
            start_weight,
            final_weight,
            start_week,
            final_week,
            period,
            scale,
            offset,
            k,
            correct_boundaries=False,
        )

        undershoot_end = (
            progression_sawtooth(
                final_week,
                start_weight,
                final_weight,
                start_week,
                final_week,
                period,
                scale,
                offset,
                k,
                correct_boundaries=False,
            )
            - final_weight
        )

        # Linear correction
        dy_dx = (overshoot_start + undershoot_end) / (final_week - start_week)
        correction = overshoot_start - dy_dx * (week - start_week)
        base_with_sinusoidal = base_with_sinusoidal + correction

    return base_with_sinusoidal


def progression_diffeq(week, start_weight, final_weight, start_week, final_week, k=0):
    """A linear/exponential progression function going through the points
    ('start_week', 'start_weight') and ('end_week', 'final_weight'), evaluated
    in 'week'.

    Parameters
    ----------
    week
        The week to evaluate the linear function at.
    start_weight
        The weight at 'start_week'.
    final_weight
        The weight at 'final_week'.
    start_week
        The number of the first week, typically 1.
    final_week
        The number of the final week, e.g. 8.
    k
        How much the function "bends". k=0 is linear, k>0 bends it.

    Returns
    -------
    weight
        The weight at 'week'.


    Examples
    -------
    >>> progression_diffeq(week = 2, start_weight = 100, final_weight = 120,
    ...                    start_week = 1, final_week = 3)
    110.0
    >>> progression_diffeq(3, 100, 140, 1, 5)
    120.0
    """
    if isinstance(week, collections.abc.Iterable):
        return list(progression_diffeq(w, start_weight, final_weight, start_week, final_week, k) for w in week)

    # assert week <= final_week
    # assert week >= start_week

    S_i = start_weight
    S_m = final_weight
    t = week
    t_i = start_week
    t_m = final_week
    assert k >= 0

    # Normalize the time
    a = (t_i - t) / (t_m - t_i)

    # Return the answer
    return (S_i - S_m) * math.exp(a * k) + S_m + a * (S_i - S_m) * math.exp(-k)


def progression_sinh(week, start_weight, final_weight, start_week, final_week, k=0):
    """A progression function similar to the hyperbolic sine, but going through
    the points ('start_week', 'start_weight') and ('end_week', 'final_weight').
    Evaluated in 'week'.

    Parameters
    ----------
    week
        The week to evaluate the linear function at.
    start_weight
        The weight at 'start_week'.
    final_weight
        The weight at 'final_week'.
    start_week
        The number of the first week, typically 1.
    final_week
        The number of the final week, e.g. 8.
    k
        How much the function "bends". k=0 is linear, k>0 bends it.

    Returns
    -------
    weight
        The weight at 'week'.

    Examples
    -------
    >>> progression_sinh(1, start_weight=3, final_weight=10, start_week=1,
    ...                     final_week=5, k=0)
    3.0
    """
    if isinstance(week, collections.abc.Iterable):
        return list(progression_sinh(w, start_weight, final_weight, start_week, final_week, k) for w in week)

    assert week <= final_week
    assert week >= start_week
    assert k >= 0

    # Normalize the time
    time_fraction = (week - start_week) / (final_week - start_week)
    assert time_fraction >= 0
    assert time_fraction <= 1

    diff_weight = (final_weight - start_weight) / 2
    mid_weight = (start_weight + final_weight) / 2

    f1 = -math.exp(-k * time_fraction)
    f2 = math.exp(k * (time_fraction - 1))

    return mid_weight + diff_weight * ((1 - time_fraction) * f1 + time_fraction * f2)


reps_to_intensity_tight = functools.partial(reps_to_intensity, slope=-3.25)
reps_to_intensity_relaxed = functools.partial(reps_to_intensity, slope=-3.75)


def wilks(lifted_kg, bodyweight_kg, gender="male"):
    """Compute Wilks points in kilograms.

    Parameters
    ----------
    lifted_kg : float
        Weight lifted.
    bodyweight_kg : float
        Bodyweight.
    gender : string, optional
        Gender of the lifter, either "male" or "female". The default is "male".

    Returns
    -------
    float
        The wilks score.

    Examples
    --------
    >>> round(wilks(500, 100, gender="male"), 2)
    304.29
    >>> round(wilks(400, 70, gender="female"), 2)
    397.94

    """
    if gender not in ("male", "female"):
        raise ValueError("`gender` must be 'male' or 'female'")

    # https://en.wikipedia.org/wiki/Wilks_Coefficient

    # Validated against https://wilkscalculator.com/kg
    # The result is not perfect. There might be a difference in coefficients
    # between the calculator and the wikipedia article. Correct to one decimal.
    if gender == "male":
        a = -216.0475144
        b = 16.2606339
        c = -0.002388645
        d = -0.00113732
        e = 7.01863e-6
        f = -1.291e-8
    if gender == "female":
        a = 594.31747775582
        b = -27.23842536447
        c = 0.82112226871
        d = -0.00930733913
        e = 4.731582e-5
        f = -9.054e-8

    x = bodyweight_kg
    coeff = 500 / (a + b * x + c * x**2 + d * x**3 + e * x**4 + f * x**5)

    return round(coeff * lifted_kg, 2)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 31 18:28:06 2019

@author: tommy
"""
import operator
from os import path

from jinja2 import Environment, FileSystemLoader

from streprogen.optimization import optimize_mealplan
import warnings


class Bunch(dict):
    def __init__(self, *args, **kwds):
        super(Bunch, self).__init__(*args, **kwds)
        self.__dict__ = self


class Mealplan:
    TEMPLATE_DIR = path.join(path.dirname(__file__), "templates")
    TEMPLATE_NAMES = {extension: "mealplan_template." + extension for extension in ["txt"]}

    def __init__(
        self,
        meals,
        dietary_constraints,
        num_meals=4,
        num_days=1,
        meal_limits=None,
        weight_price=0.1,
        weight_nutrients=2.0,
        weight_meal_sizes=0.75,
    ):
        """Initialize a new program.

        Parameters
        ----------
        meals
            A list of Meal instances.

        dietary_constraints
            A dict of dietary constraints. The keys can be 'kcal', 'protein', 'fat' or
            'carbs'. The values must be (low, high) limits of the macronutrients, or
            (None, high) to only set an upper limit. Example: {'kcal' : (1600, 1800)}

        num_meals
            The number of meals per day.

        num_days
            The number of days in the meal plan.

        meal_limits
            A dictionary with keys equal to meal names. The values must be (low, high)
            and limit the number of times a food can be used. Use (None, high) to only
            set an upper limit. Example: To have the meal 'bread' at most twice in a
            multi-day meal plan, use {'bread' : (None, 2)}.

        weight_price
            Positive number weighting the overall price of the meal plan.
            A higher value weights this objective higher.
            This value must be set in relation to the other weights.

        weight_nutrients
            Positive number weighting the objective of getting as close as possible to
            satisfying the dietary constraints. A higher value weights this objective
            higher.  This value must be set in relation to the other weights.

        weight_meal_sizes
            Positive number weighting the goal of having equal meal sizes.
            A higher value weights this objective higher.
            This value must be set in relation to the other weights.

        """
        assert num_meals >= 1
        assert num_days >= 1

        self.meals = meals
        self.dietary_constraints = dietary_constraints
        self.num_meals = num_meals
        self.num_days = num_days

        self.weight_price = weight_price
        self.weight_nutrients = weight_nutrients
        self.weight_meal_sizes = weight_meal_sizes

        if meal_limits is None:
            self.meal_limits = dict()
        else:
            self.meal_limits = meal_limits

        # Add None keys if no keys are present
        for meal in self.meals:
            if meal.name not in set(self.meal_limits.keys()):
                self.meal_limits[meal.name] = (None, None)

        assert len(self.meals) == len(self.meal_limits)

        meal_names = set([meal.name for meal in self.meals])
        for key in self.meal_limits.keys():
            if key not in meal_names:
                msg = "`meal_limits` has a key `{}` which is not a meal name."
                warnings.warn(msg.format(key))

        self._rendered = False
        self._set_jinja2_enviroment()

    def render(self, time_limit_secs=5, epsilon=1e-3, params=None):
        """Render the meal plan to perform the calculations.

        Parameters
        ----------
        time_limit_secs
            Solver time limit.

        epsilon
            Small number used in optimization routine.

        params
            Additional low-level parameters used in optimization. See source code.

        """
        limits = [self.meal_limits[meal.name] for meal in self.meals]

        x, optimization_results = optimize_mealplan(
            self.meals,
            self.dietary_constraints,
            meals_limits=limits,
            num_days=self.num_days,
            num_meals=self.num_meals,
            time_limit_secs=time_limit_secs,
            epsilon=epsilon,
            weight_price=self.weight_price,
            weight_nutrients=self.weight_nutrients,
            weight_range=self.weight_meal_sizes,
            params=params,
        )

        self.x = x
        self.optimization_results = optimization_results
        self._rendered = True

        # Parse the results
        self.results = Bunch()
        for attr in ["price", "protein", "fat", "carbs", "kcal"]:
            self.results[attr] = []
        self.results["pretty"] = []

        num_meals = len(x)
        num_days = len(x[0])

        # Store used meals
        self.results["meals"] = set()
        for i in range(len(x)):
            for j in range(len(x[i])):
                if x[i][j] > 0:
                    self.results["meals"].add(self.meals[i])

        self.results["meals"] = list(sorted(self.results["meals"], key=operator.attrgetter("name")))

        def format_qntity(qnty):
            """Format a number."""
            qnty = round(qnty, 1)
            if qnty % 1 == 0:
                qnty = int(qnty)
            return qnty

        for day_num in range(num_days):
            x_day = [x[i][day_num] for i in range(num_meals)]

            result = [(meal, qnty) for (meal, qnty) in zip(self.meals, x_day) if qnty > 0]
            result = sorted(result, key=lambda r: r[0].carbs * r[1], reverse=True)

            self.results[day_num] = dict()
            for attr in ["price", "protein", "fat", "carbs", "kcal"]:
                self.results[day_num][attr] = [getattr(m, attr) * q for (m, q) in result]
                self.results[attr].append(sum(self.results[day_num][attr]))

            # Heuristics to get more carbohydrates earlier in the day
            result = [(meal, format_qntity(qnty)) for (meal, qnty) in result]
            result = [(meal, str(qnty).ljust(3)) for (meal, qnty) in result]
            self.results["pretty"].append(result)

    def _set_jinja2_enviroment(self):
        """
        Set up the jinja2 environment.
        """

        template_loader = FileSystemLoader(searchpath=self.TEMPLATE_DIR)
        env = Environment(loader=template_loader, trim_blocks=True, lstrip_blocks=True)
        # env.globals.update(chunker=chunker, enumerate=enumerate, str=str)

        self.jinja2_environment = env

    def to_txt(self, verbose=False):
        """Write the program information to text, which can be printed in a terminal.

        Parameters
        ----------
        verbose
            If True, more information is shown.

        Returns
        -------
        string
            Meal plan as text.
        """

        env = self.jinja2_environment
        template = env.get_template(self.TEMPLATE_NAMES["txt"])
        return template.render(mealplan=self, results=self.results, verbose=verbose)

    def to_html(self, verbose=False):
        # TODO: Implement this.
        raise NotImplementedError("Not implemented yet.")

    def to_tex(self, verbose=False):
        # TODO: Implement this.
        raise NotImplementedError("Not implemented yet.")

    def __str__(self):
        """
        String formatting for readable human output.
        """
        return self.to_txt()
